#include "params.h"

const int nAsym = 7;
const float gAsym[] = {1.0, 4.0, 2.0, 4.0, 2.0, 4.0, 1.0};
const float coefAsym = 1.0 / 18.0;

extern LPSTR iniFile;

void TPeakFitDialog::IDPeakFitPeakSearch(RTMessage)
{
	if(Data[0] == NULL) return;

	PeakData Peak;
	float f;

	PTPeakSearchDialog pTPSD = new TPeakSearchDialog(this, DIALOG_PEAKSEARCH, Frame, &Peak);
	TheParent->AutoRedraw = FALSE;
	if(GetModule()->ExecDialog(pTPSD) == IDCANCEL) return;
	TheParent->AutoRedraw = TRUE;
	if(Peak.nPeak < 1) return;

	if(Peak.nPeak > MAX_FITPEAK)
	{
		MessageBox(HWindow, "Too many peaks. Select OK to continue",
			"Warning!!", MB_OK);
		 pPeak->nPeak = MAX_FITPEAK;
	}
	else pPeak->nPeak = Peak.nPeak;
	BOOL *flag = (BOOL *)farmalloc(Peak.nPeak * sizeof(BOOL));
	if(flag == NULL)
	{
		MessageBox(HWindow, "Can't allocate memory (TPeakFitDialog::IDPeakFitPeakSearch flag)",
			"Warning!!", MB_OK);
		pPeak->nPeak = 0;
		return;
	}
	int i, j;
	for(i = 0 ; i < pPeak->nPeak ; i++)
			flag[i] = TRUE;
	float Q1, Q2, d, diff, mindiff;
	const float QCriteria = 0.05;
	int idx;
	int count = 0;
	Real BK = pPeak->Back[0] = pPeak->Data->OrgRange.Ymin() * 0.9;
	for(i = 0 ; i < pPeak->nPeak ; i++)
	{
		if(flag[i] == FALSE) continue;
		Q1 = Peak.angle[i];
		if(Q1 <= 0.0 || Q1 >= 180.0) continue;
		d = sin(Q1*0.5*DegToRad) / pPeak->ramda * 2.0;
		Q2 = 2.0*RadToDeg * asin(pPeak->ramda2*0.5 * d);

		mindiff = 1e30;
		idx = -1;
		for(j = i+1 ; j < pPeak->nPeak ; j++)
		{
			diff = fabs(Peak.angle[j] - Q2);
			if(diff < mindiff)
			{
				mindiff = diff;
				idx = j;
			}
		}
		if(idx > 0 && mindiff < QCriteria &&
			Peak.intensity[i] > Peak.intensity[idx]) flag[idx] = FALSE;

		if(Q1 > 70.0) f = 1.0;
		else f = 1.0 / (1.0 + pPeak->IRatio * (1.0 - Q1 / 70.0));
		pPeak->Angle[count] = Q1;
		pPeak->Height[count] = Peak.intensity[i] * f - BK;
		pPeak->Gfwhm[count] = Peak.fwhm[i] * f;
		pPeak->Lfwhm[count] = Peak.fwhm[i] * f;
		pPeak->L_GfwhmRatio[count] = 1.0;
		pPeak->GFrac[count] = 0.5;
		pPeak->Aniso[count] = 0.0;
		count++;
	}
	pPeak->nPeak = count;
	farfree(flag);
	UpdateDialog();
}

PeakFitData::PeakFitData(PDataStruc pds)
{
	if(pds != NULL) strcpy(Title, pds->Title);
	if(pds != NULL) strcpy(FileName, pds->FileName);
	nPeak = 0;
	curPeak = 0;
	isAutoRedraw = TRUE;
	Data = pds;
	Back[0] = 0.0;
	Back[1] = 0.0;
	Back[2] = 0.0;
	Back[3] = 0.0;
	Back[4] = 0.0;
	idQRange = TRUE;
	idIRatio = TRUE;

	int i;
	for(i = 0 ; i < MAX_FITPEAK ; i++)
	{
		Angle[i] = 30.0;
		GFrac[i] = 0.5;
		Height[i] = 1000;
		Gfwhm[i] = 0.16;
		Lfwhm[i] = 0.16;
		L_GfwhmRatio[i] = 1.0;
		Aniso[i] = 0.0;
		idAngle[i] = TRUE;
		idHeight[i] = TRUE;
		idGFrac[i] = TRUE;
		idGfwhm[i] = TRUE;
		idLfwhm[i] = TRUE;
		idL_GfwhmRatio[i] = TRUE;
		idAniso[i] = TRUE;
	}
	strcpy(OptProgram, "xPeakFit.exe");
	strcpy(OptMethod, "Gauss");
	IRatio = 0.5;
	QRange = 5.0;
	ramda = 1.540562;
	ramda2 = 1.544390;
	Dumping = 0.3;
	nCycle = 10;
	EPS = 1.0e-6;
	idBack[0] = TRUE;
	idBack[1] = TRUE;
	idBack[2] = TRUE;
	idBack[3] = FALSE;
	idBack[4] = FALSE;
//	GetProfile(iniFile, "PeakFit");
}

PeakFitData::~PeakFitData()
{
}

BOOL PeakFitData::GetProfile(char *iniFile, char *section)
{
	char buff[BUF_LEN];

	GetPrivateProfileString(section, "idBack0", "1",
		buff, BUF_LEN, iniFile);
	if(*buff == '1') idBack[0] = TRUE;
	else idBack[0] = FALSE;
	GetPrivateProfileString(section, "idBack1", "1",
		buff, BUF_LEN, iniFile);
	if(*buff == '1') idBack[1] = TRUE;
	else idBack[1] = FALSE;
	GetPrivateProfileString(section, "idBack2", "0",
		buff, BUF_LEN, iniFile);
	if(*buff == '1') idBack[2] = TRUE;
	else idBack[2] = FALSE;
	GetPrivateProfileString(section, "idBack3", "0",
		buff, BUF_LEN, iniFile);
	if(*buff == '1') idBack[3] = TRUE;
	else idBack[3] = FALSE;
	GetPrivateProfileString(section, "idBack4", "0",
		buff, BUF_LEN, iniFile);
	if(*buff == '1') idBack[4] = TRUE;
	else idBack[4] = FALSE;
//	idBack[4] = FALSE;

	GetPrivateProfileString(section, "IRatio", "0.5",
		buff, BUF_LEN, iniFile);
	IRatio = atof(buff);
	GetPrivateProfileString(section, "QRange", "5.0",
		buff, BUF_LEN, iniFile);
	QRange = atof(buff);
	GetPrivateProfileString(section, "ramda", "1.540562",
		buff, BUF_LEN, iniFile);
	ramda = atof(buff);
	GetPrivateProfileString(section, "ramda2", "1.544390",
		buff, BUF_LEN, iniFile);
	ramda2 = atof(buff);

	if(*buff == '1') idBack[4] = TRUE;
	else idBack[4] = FALSE;

	int i;
	BOOL id;
	GetPrivateProfileString(section, "idAngle", "1",
		buff, BUF_LEN, iniFile);
	if(*buff == '1') id = TRUE;
	else id = FALSE;
	for(i = 0 ; i < MAX_FITPEAK ; i++)
		idAngle[i] = id;
	GetPrivateProfileString(section, "idHeight", "1",
		buff, BUF_LEN, iniFile);
	if(*buff == '1') id = TRUE;
	else id = FALSE;
	for(i = 0 ; i < MAX_FITPEAK ; i++)
		idHeight[i] = id;
	GetPrivateProfileString(section, "idGFrac", "1",
		buff, BUF_LEN, iniFile);
	if(*buff == '1') id = TRUE;
	else id = FALSE;
	for(i = 0 ; i < MAX_FITPEAK ; i++)
		idGFrac[i] = id;
	GetPrivateProfileString(section, "idGfwhm", "1",
		buff, BUF_LEN, iniFile);
	if(*buff == '1') id = TRUE;
	else id = FALSE;
	for(i = 0 ; i < MAX_FITPEAK ; i++)
		idGfwhm[i] = id;
	GetPrivateProfileString(section, "idLfwhm", "1",
		buff, BUF_LEN, iniFile);
	if(*buff == '1') id = TRUE;
	else id = FALSE;
	for(i = 0 ; i < MAX_FITPEAK ; i++)
		idLfwhm[i] = id;
	GetPrivateProfileString(section, "idrfwhm", "1",
		buff, BUF_LEN, iniFile);
	if(*buff == '1') id = TRUE;
	else id = FALSE;
	for(i = 0 ; i < MAX_FITPEAK ; i++)
		idL_GfwhmRatio[i] = id;
	GetPrivateProfileString(section, "idAniso", "1",
		buff, BUF_LEN, iniFile);
	if(*buff == '1') id = TRUE;
	else id = FALSE;
	for(i = 0 ; i < MAX_FITPEAK ; i++)
		idAniso[i] = id;

	GetPrivateProfileString(section, "Dumping", "0.3",
		buff, BUF_LEN, iniFile);
	Dumping = atof(buff);
	GetPrivateProfileString(section, "EPS", "1.0e-6",
		buff, BUF_LEN, iniFile);
	EPS = atof(buff);
	GetPrivateProfileString(section, "nCycle", "10",
		buff, BUF_LEN, iniFile);
	nCycle = atoi(buff);
	GetPrivateProfileString(section, "OptProgram", "xPeakFit.exe",
		OptProgram, BUF_LEN, iniFile);
	GetPrivateProfileString(section, "OptMethod", "Marq",
		OptMethod, BUF_LEN, iniFile);
	return (TRUE);
}

BOOL PeakFitData::Save(char *file, PDataStruc Data)
{
	FILE *fp;
	char str[LEN_STRING];

	if(nPeak < 1) return(FALSE);
	if(strchr(file, '.') == NULL) strcat(file, ".pfc");
	if((fp = fopen(file, "r")) != NULL)
	{
		fclose(fp);
		sprintf(str, "Overwrite %s\n OK?", file);
		if(MessageBox(TheParent->HWindow, str, "Confirm!!", MB_OKCANCEL) == IDCANCEL)
				return (FALSE);
	}
	if((fp = FileOpen(file, "w")) == NULL) return(FALSE);
	strcpy(Title, Data->Title);
	strcpy(FileName, Data->FileName);
	fprintf(fp, "%s\n", Data->Title);
	fprintf(fp, "%s\n", Data->FileName);
	fprintf(fp, "%s\n", OptProgram);
	fprintf(fp, "%s\n", OptMethod);
	fprintf(fp, "%d %g %g\n", nCycle, Dumping, EPS);
	fprintf(fp, "%g %g %g %g\n", ramda, ramda2, QRange, IRatio);
	fprintf(fp, "%g %g\n", CalcX0, CalcX1);
	fprintf(fp, "%g %g\n", OptX0, OptX1);
	fprintf(fp, "  %g  %d  0.0\n", Back[0], idBack[0]);
	fprintf(fp, "  %g  %d  0.0\n", Back[1], idBack[1]);
	fprintf(fp, "  %g  %d  0.0\n", Back[2], idBack[2]);
	fprintf(fp, "  %g  %d  0.0\n", Back[3], idBack[3]);
	fprintf(fp, "  %g  %d  0.0\n", Back[4], idBack[4]);
	fprintf(fp, "%d\n", nPeak);
	int i;
	for(i = 0 ; i < nPeak ; i++)
	{
		fprintf(fp, "  %g  %d  0.0\n", Angle[i], idAngle[i]);
		fprintf(fp, "  %g  %d  0.0\n", Height[i], idHeight[i]);
		fprintf(fp, "  %g  %d  0.0\n", GFrac[i], idGFrac[i]);
		fprintf(fp, "  %g  %d  0.0\n", Gfwhm[i], idGfwhm[i]);
		fprintf(fp, "  %g  %d  0.0\n", Lfwhm[i], idLfwhm[i]);
		fprintf(fp, "  %g  %d  0.0\n", L_GfwhmRatio[i], idL_GfwhmRatio[i]);
		fprintf(fp, "  %g  %d  0.0\n", Aniso[i], idAniso[i]);
	}
	fprintf(fp, "\n%ld\n", Data->datan);
	Long l;
	for(l = 0 ; l < Data->datan ; l++)
		fprintf(fp, "  %g  %g\n", Data->XVal(l), Data->YVal(l));
	fclose(fp);
	return (TRUE);
}

BOOL PeakFitData::Read(char *file, PFrameStruc pfs)
{
	Long datan, i;
	Real x, y, x0, y0, x1, y1;

	FILE *fp;
	if((fp = FileOpen(file, "r")) == NULL) return(FALSE);
	BOOL ret = Read(fp);
	fscanf(fp, "%ld", &datan);

	fscanf(fp, "%g %g", &x0, &y0);
	fscanf(fp, "%g %g", &x1, &y1);
	pfs->AddData(x0, x1-x0, datan);
	DataStruc *data = pfs->GetData(0);
	data->YSet(0, y0);
	data->YSet(1, y1);
	strcpy(data->Title, Title);
	strcpy(data->FileName, FileName);
	for(i = 2 ; i < datan ; i++)
	{
		fscanf(fp, "%g %g", &x, &y);
		data->YSet(i, y);
	}
	fclose(fp);
	pfs->Torelance.SetYmin(0.3);
	pfs->SetFullRange();
//	pfs->SetViewYmin(0.0);
	Data = data;
	strcpy(pfs->xString.str, "CuKa 2Theta / degree");
	strcpy(pfs->yString.str, "Intensity / cps");
	return ret;
}

BOOL PeakFitData::Read(char *file, PDataStruc)
{
	FILE *fp;
	if((fp = FileOpen(file, "r")) == NULL) return(FALSE);
	BOOL ret = Read(fp);
	fclose(fp);
	return ret;
}

BOOL PeakFitData::Read(FILE *fp)
{
	char str[LEN_STRING], *cp;
	float val;

	fgets(Title, LEN_STRING, fp);
	cp = Title + strlen(Title) - 1;
	if(*cp == '\n' || *cp == '\r') *cp = NULL;
	fgets(FileName, LEN_STRING, fp);
	cp = FileName + strlen(FileName) - 1;
	if(*cp == '\n' || *cp == '\r') *cp = NULL;
	fscanf(fp, "%s", OptProgram);
	fscanf(fp, "%s", OptMethod);
	fscanf(fp, "%d %g %g", &nCycle, &Dumping, &EPS);
	fscanf(fp, "%g %g %g %g", &ramda, &ramda2, &QRange, &IRatio);
	fscanf(fp, "%g %g", &CalcX0, &CalcX1);
	fscanf(fp, "%g %g", &OptX0, &OptX1);
	fscanf(fp, "%g %d %g", &Back[0], &idBack[0], &val);
	fscanf(fp, "%g %d %g", &Back[1], &idBack[1], &val);
	fscanf(fp, "%g %d %g", &Back[2], &idBack[2], &val);
	fscanf(fp, "%g %d %g", &Back[3], &idBack[3], &val);
	fscanf(fp, "%g %d %g", &Back[4], &idBack[4], &val);
	fscanf(fp, "%d", &nPeak);
	int i;
	for(i = 0 ; i < nPeak ; i++)
	{
		fscanf(fp, "%g %d %g", &(Angle[i]), &(idAngle[i]), &val);
		fscanf(fp, "%g %d %g", &(Height[i]), &(idHeight[i]), &val);
		fscanf(fp, "%g %d %g", &(GFrac[i]), &(idGFrac[i]), &val);
		fscanf(fp, "%g %d %g", &(Gfwhm[i]), &(idGfwhm[i]), &val);
		fscanf(fp, "%g %d %g", &(Lfwhm[i]), &(idLfwhm[i]), &val);
		fscanf(fp, "%g %d %g", &(L_GfwhmRatio[i]), &(idL_GfwhmRatio[i]), &val);
		fscanf(fp, "%g %d %g", &(Aniso[i]), &(idAniso[i]), &val);
	}
	return (TRUE);
}

void PeakFitData::Plot(HWND HWindow, PDeviceInfo pDevInf, PFrameStruc Frame)
{
	if(nPeak < 1 || Data == NULL) return;

	PDataStruc dt = Data;
	Long datan = dt->datan;
	Real f1, f2;
	Real Qmin, Qmax, qw, Q, xx1, xx2, Q1, Q2, d;
	Long i0, i1, i, j, k;
	Real Qa;
	Real val, CotQ1, CotQ2, coeL;
	const Real coeG = 2.772588722;

	if(dt->xstep == 0.0)
	{
		MessageBox(HWindow, "Zero xstep (PeakFitData::Plot dt->xzero)",
			"Warning!!", MB_OK);
		return;
	}
	DataStruc APeak(Frame, 0, datan);
	DataStruc BackG(Frame, 0, datan);
	DataStruc Total(Frame, 0, datan);
	APeak.xstart = dt->xstart;
	APeak.xstep  = dt->xstep;
	BackG.xstart  = dt->xstart;
	BackG.xstep   = dt->xstep;
	Total.xstart = dt->xstart;
	Total.xstep  = dt->xstep;
	Total.LineStyle.rgbColor = RGB_RED;
	BackG.LineStyle.rgbColor = RGB_CYAN;
	if(QRange < 0.1) QRange = 0.1;
	for(i = 0 ; i < datan ; i++)
	{
		Q = BackG.XVal(i);
		val = Back[0] + (Back[1] + (Back[2] + (Back[3] + Back[4]*Q) * Q) * Q) * Q;
		BackG.YSet(i, val);
	}
	BackG.Plot(HWindow, pDevInf, Frame->GetViewXmin(), Frame->GetViewXmax());
		//i0, i1);
	for(i = 0 ; i < datan ; i++)
	{
		Total.YSet(i, BackG.YVal(i));
	}
	Real Lf, Gf, rf, fj[7];
	for(i = 0 ; i < 7 ; i++)
	{
		fj[i] = (double)i / (double)(nAsym-1);
		fj[i] = fj[i] * fj[i];
	}
	for(i = 0 ; i < nPeak ; i++)
	{
		Lf = Lfwhm[i];
		Gf = Gfwhm[i];
		if(Gf >= Lf) qw = Gf * QRange;
		else qw = Lf * QRange;
		Q1 = Angle[i];
		if(Q1 <= 0.0 || Q1 >= 180.0) continue;
		d = sin(Q1*0.5*DegToRad) / ramda * 2.0;
		Q2 = 2.0*RadToDeg * asin(ramda2*0.5 * d);
		Qmin = Angle[i] - qw;
		Qmax = Q2 + qw;
		i0 = (Long)((Qmin - dt->xstart) / dt->xstep);
		i1 = (Long)((Qmax - dt->xstart) / dt->xstep);
		if(i0 < 0) i0 = 0;
		if(i0 > datan) continue;
		if(i1 > datan) i1 = datan;
		if(i1 < 0) continue;
		rf = Gf / Lf;
		coeL = rf * 4.0;
		for(j = i0 ; j < i1 ; j++)
			APeak.YSet(j, 0);
		CotQ1 = 1.0 / tan(Q1 * DegToRad);
		CotQ2 = 1.0 / tan(Q2 * DegToRad);
		for(j = i0 ; j < i1 ; j++)
		{
			Q = dt->xstart + dt->xstep * (float)j;
			f1 = 0.0;
			for(k = 0 ; k < nAsym ; k++)
			{
				Qa = Q-Q1 + fj[k]*Aniso[i] * CotQ1;
				xx1 = Qa / Gfwhm[i];
				f2 = (GFrac[i] * exp(-coeG * xx1*xx1)
						+(1.0-GFrac[i]) / (1.0 + coeL * xx1*xx1));
				Qa = Q-Q2 + fj[k]*Aniso[i] * CotQ2;
				xx2 = Qa / Gfwhm[i];
				f2 = f2 + IRatio * (GFrac[i] * exp(-coeG * xx2*xx2)
						+(1.0-GFrac[i]) / (1.0 + coeL * xx2*xx2));
				f1 = f1 + f2 * gAsym[k];
			}
			f1 = f1 * coefAsym * Height[i];
			APeak.YSet(j, f1);
			Total.YSet(j, Total.YVal(j)+f1);
		}
		if(i == curPeak)
			APeak.LineStyle.rgbColor = RGB_BLUE;
		else APeak.LineStyle.rgbColor = RGB_GREEN;
			APeak.Plot(HWindow, pDevInf, i0, i1);
	}
	Total.Plot(HWindow, pDevInf);
//        FrameStruc errfrm(40, 140, 0, 20);
//        errfrm.yString.direction = Y_DIRECTION;
	DataStruc yerror(Frame, 0, datan);
	yerror.xstart = dt->xstart;
	yerror.xstep = dt->xstep;
	yerror.LineStyle.rgbColor = RGB_MAGENTA;
	float yshift = (Frame->GetViewYmax() + Frame->GetViewYmin()) / 2.0;
	for(i = 0 ; i < datan ; i++)
		yerror.YSet(i, dt->YVal(i) - Total.YVal(i) + yshift);
	yerror.ViewRange.SetMinMax(Frame->GetViewXmin(),
			-fabs(Frame->GetViewYmax() * 0.1),
	Frame->GetViewXmax(),
			fabs(Frame->GetViewYmax() * 0.1));
/*
		yerror.ViewRange.xmin = Frame->ViewRange.xmin;
		yerror.ViewRange.xmax = Frame->ViewRange.xmax;
		yerror.ViewRange.ymin = -fabs(Frame->ViewRange.ymax * 0.1);
		yerror.ViewRange.ymax = -yerror.ViewRange.ymin;
*/
//        errfrm.SetDeviceCoord(pDevInf);
//        errfrm.Plot(pDevInf);
	yerror.Plot(HWindow, pDevInf);
}

void TMyWindow::CMDataPeakFit(RTMessage)
{
	PFrameStruc pfs;

	if(nFrame < 1) return;
	if((pfs = GetSelectedFrame(SelectedObj)) == NULL)
			pfs = Frame[0];
	if(pfs->GetDataNumber() < 1) return;
	PTPeakFitDialog pTPFD = new TPeakFitDialog(&FitPeak, pfs, this, DIALOG_PEAKFIT);
	GetModule()->MakeWindow(pTPFD);
	pTPFD->Show(SW_SHOW);
}

TPeakFitDialog::TPeakFitDialog(PPeakFitData ppfd, PFrameStruc pfs, PTWindowsObject AParent,
	int ResourceId, PTModule TheModule) : TDialog(AParent, ResourceId, TheModule)
{
	int i;

	if(pfs->GetDataNumber() < 1) return;
	if(ppfd->nPeak <= 0)
	{
		ppfd->GetProfile(iniFile, "PeakFit");
		ppfd->idBack[4] = FALSE;
	}

	Frame = pfs;
	pPeak = ppfd;
	memcpy(&back, ppfd, sizeof(PeakFitData));
	nData = pfs->GetDataNumber();
	for(i = 0 ; i < pfs->GetDataNumber() ; i++)
		Data[i] = pfs->GetData(i);
	curData = 0;
	ppfd->OptX0 = ppfd->CalcX0 = pfs->GetData(0)->OrgRange.Xmin();
	ppfd->OptX1 = ppfd->CalcX1 = pfs->GetData(0)->OrgRange.Xmax();
	pTDataList 		= new TComboBox(this, ID_PEAKFITDATANUMBER, LEN_STRING);
	pTGaussList 	= new TComboBox(this, ID_PEAKFITGAUSSNUMBER, LEN_STRING);
	pTIRatio		= new TFloatEdit(this, ID_PEAKFITIRATIO);
	pTAutoRedraw	= new TCheckBox(this, ID_PEAKFITAUTOREDRAW, NULL);
	pTAngle			= new TFloatEdit(this, ID_PEAKFITANGLE);
	pTGFrac			= new TFloatEdit(this, ID_PEAKFITGAUSSFRAC);
	pTHeight		= new TFloatEdit(this, ID_PEAKFITHEIGHT);
	pTGfwhm			= new TFloatEdit(this, ID_PEAKFITGAUSSFWHM);
	pTLfwhm			= new TFloatEdit(this, ID_PEAKFITLORENTZFWHM);
	pTAngleScroll	= new TScrollBar(this, ID_PEAKFITANGLESCROLL);
	pTGFracScroll	= new TScrollBar(this, ID_PEAKFITGFRACSCROLL);
	pTHeightScroll	= new TScrollBar(this, ID_PEAKFITHEIGHTSCROLL);
	pTGfwhmScroll	= new TScrollBar(this, ID_PEAKFITGAUSSFWHMSCROLL);
	pTLfwhmScroll	= new TScrollBar(this, ID_PEAKFITLORENTZFWHMSCROLL);
	pTOk			= new TButton(this, IDOK);
	pTCancel		= new TButton(this, IDCANCEL);
	pTView			= new TButton(this, ID_VIEW);
	pTAdd			= new TButton(this, ID_PEAKFITADD);
	pTDelete		= new TButton(this, ID_PEAKFITDELETE);
	pTfwhmRatio		= new TFloatEdit(this, ID_PEAKFITFWHMRATIO);
	pTQRange		= new TFloatEdit(this, ID_PEAKFITQRANGE);
	pTfwhmRatioScroll = new TScrollBar(this, ID_PEAKFITFWHMRATIOSCROLL);
	pTBackA0		= new TFloatEdit(this, ID_PEAKFITA0);
	pTBackA0Scroll	= new TScrollBar(this, ID_PEAKFITA0SCROLL);
	pTBackA1		= new TFloatEdit(this, ID_PEAKFITA1);
	pTBackA1Scroll	= new TScrollBar(this, ID_PEAKFITA1SCROLL);
	pTBackA2		= new TFloatEdit(this, ID_PEAKFITA2);
	pTBackA2Scroll	= new TScrollBar(this, ID_PEAKFITA2SCROLL);
	pTPeakSearch	= new TButton(this, ID_PEAKFITPEAKSEARCH);
	pTRamda			= new TFloatEdit(this, ID_PEAKFITRAMDA);
	pTRamda2		= new TFloatEdit(this, ID_PEAKFITRAMDA2);
	pTAnisotropic	= new TFloatEdit(this, ID_PEAKFITANISOTROPIC);
	pTAnisotropicScroll	= new TScrollBar(this, ID_PEAKFITANISOTROPICSCROLL);

	TransferBuffer = NULL;
}

TPeakFitDialog::~TPeakFitDialog()
{
	delete pTDataList;
	delete pTGaussList;
	delete pTIRatio;
	delete pTAutoRedraw;
	delete pTAngle;
	delete pTGFrac;
	delete pTHeight;
	delete pTGfwhm;
	delete pTLfwhm;
	delete pTAngleScroll;
	delete pTGFracScroll;
	delete pTHeightScroll;
	delete pTGfwhmScroll;
	delete pTLfwhmScroll;
	delete pTOk;
	delete pTCancel;
	delete pTView;
	delete pTAdd;
	delete pTDelete;
	delete pTfwhmRatio;
	delete pTQRange;
	delete pTfwhmRatioScroll;
	delete pTBackA0;
	delete pTBackA0Scroll;
	delete pTBackA1;
	delete pTBackA1Scroll;
	delete pTBackA2;
	delete pTBackA2Scroll;
	delete pTRamda;
	delete pTRamda2;
	delete pTPeakSearch;
	delete pTAnisotropic;
	delete pTAnisotropicScroll;
}

void TPeakFitDialog::SetupWindow(void)
{
	BOOL fback;

	fback = pPeak->isAutoRedraw;
	pPeak->isAutoRedraw = FALSE;

	TDialog::SetupWindow();
	UpdateDialog();
	pPeak->isAutoRedraw = fback;
	if(pPeak->isAutoRedraw) pTAutoRedraw->Check();
	else pTAutoRedraw->Uncheck();
	if(pPeak->nPeak > 0)
	{
		TMessage Msg;
		Msg.LP.Hi = LBN_SELCHANGE;
		IDGaussList(Msg);
	}
}

void TPeakFitDialog::UpdateDialog(void)
{
	char str[LEN_STRING];
	int i;

	pPeak->Data = Data[curData];
	pTDataList->ClearList();
	for(i = 0 ; i < nData ; i++)
	{
		if(Data[i] == NULL) continue;
		if(Data[i]->Title[0] == NULL)
			sprintf(str, "%2d:%s", i+1, Data[i]->FileName);
		else	sprintf(str, "%2d:%s", i+1, Data[i]->Title);
		pTDataList->AddString(str);
	}
	pTDataList->SetSelIndex(curData);

	UpdateGaussList();
	pTIRatio->SetValue(pPeak->IRatio);
	pTAngle->SetValue(pPeak->Angle[pPeak->curPeak]);
	pTGFrac->SetValue(pPeak->GFrac[pPeak->curPeak]);
	pTHeight->SetValue(pPeak->Height[pPeak->curPeak]);
	pTGfwhm->SetValue(pPeak->Gfwhm[pPeak->curPeak]);
	pTLfwhm->SetValue(pPeak->Lfwhm[pPeak->curPeak]);
	pTfwhmRatio->SetValue(pPeak->L_GfwhmRatio[pPeak->curPeak]);
	pTAnisotropic->SetValue(pPeak->Aniso[pPeak->curPeak]);
	pTQRange->SetValue(pPeak->QRange);
	pTIRatio->SetValue(pPeak->IRatio);
	pTBackA0->SetValue(pPeak->Back[0]);
	pTBackA1->SetValue(pPeak->Back[1]);
	pTBackA2->SetValue(pPeak->Back[2]);
	pTRamda->SetValue(pPeak->ramda);
	pTRamda2->SetValue(pPeak->ramda2);
	pTAngleScroll->LineMagnitude = 1;
	pTAngleScroll->PageMagnitude = 10;
	pTAngleScroll->SetRange(0, 10000);
	pTGFracScroll->LineMagnitude = 10;
	pTGFracScroll->PageMagnitude = 1000;
	pTGFracScroll->SetRange(0, 10000);
	pTHeightScroll->LineMagnitude = 50;
	pTHeightScroll->PageMagnitude = 250;
	pTHeightScroll->SetRange(0, 10000);
	pTGfwhmScroll->LineMagnitude = 10;
	pTGfwhmScroll->PageMagnitude = 1000;
	pTGfwhmScroll->SetRange(0, 10000);
	pTGfwhmScroll->SetPosition(5000);
	pTLfwhmScroll->LineMagnitude = 10;
	pTLfwhmScroll->PageMagnitude = 1000;
	pTLfwhmScroll->SetRange(0, 10000);
	pTLfwhmScroll->SetPosition(5000);
	pTfwhmRatioScroll->LineMagnitude = 10;
	pTfwhmRatioScroll->PageMagnitude = 1000;
	pTfwhmRatioScroll->SetRange(0, 10000);
	pTBackA0Scroll->LineMagnitude = 10;
	pTBackA0Scroll->PageMagnitude = 1000;
	pTBackA0Scroll->SetRange(0, 10000);
	pTBackA1Scroll->LineMagnitude = 10;
	pTBackA1Scroll->PageMagnitude = 1000;
	pTBackA1Scroll->SetRange(0, 10000);
	pTBackA2Scroll->LineMagnitude = 10;
	pTBackA2Scroll->PageMagnitude = 1000;
	pTBackA2Scroll->SetRange(0, 10000);
	pTAnisotropicScroll->LineMagnitude = 10;
	pTAnisotropicScroll->PageMagnitude = 1000;
	pTAnisotropicScroll->SetRange(0, 10000);
}

void TPeakFitDialog::UpdateGaussList(void)
{
	int i;
	char str[LEN_STRING];

	pTGaussList->ClearList();
	for(i = 0 ; i < pPeak->nPeak ; i++)
	{
		sprintf(str, "%2d:%gdeg. %gcps", i+1, pPeak->Angle[i], pPeak->Height[i]);
		pTGaussList->AddString(str);
	}
	pTGaussList->SetSelIndex(pPeak->curPeak);

	if(pTAutoRedraw->GetCheck() == BF_CHECKED)
		InvalidateRect(Parent->HWindow, NULL, TRUE);
}

void TPeakFitDialog::IDDataList(RTMessage Msg)
{
	if(Msg.LP.Hi != LBN_SELCHANGE) return;
	curData = pTDataList->GetSelIndex();
	if(0 <= curData && curData < nData) UpdateDialog();
		else curData = 0;
	pPeak->Data = Data[curData];
	pPeak->OptX0 = pPeak->CalcX0 = pPeak->Data->OrgRange.Xmin();
	pPeak->OptX1 = pPeak->CalcX1 = pPeak->Data->OrgRange.Xmax();
}

void TPeakFitDialog::IDGaussList(RTMessage Msg)
{
	if(Msg.LP.Hi != LBN_SELCHANGE) return;
	pPeak->curPeak = pTGaussList->GetSelIndex();
	if(0 <= pPeak->curPeak && pPeak->curPeak < pPeak->nPeak) UpdateDialog();
	else pPeak->curPeak = 0;
}

void TPeakFitDialog::IDAutoRedraw(RTMessage)
{
	if(pTAutoRedraw->GetCheck() == BF_CHECKED)
	{
		pPeak->isAutoRedraw = TRUE;
		InvalidateRect(Parent->HWindow, NULL, TRUE);
	}
	else pPeak->isAutoRedraw = FALSE;
}

void TPeakFitDialog::IDPeakFitRamda(RTMessage Msg)
{
	if(Msg.LP.Hi != EN_UPDATE) return;
	float val = pTRamda->GetValue();
	if(!pTRamda->isValid) return;
	if(val <= 0.0) val = 1.5405;
	pPeak->ramda = val;
	UpdateGaussList();
}

void TPeakFitDialog::IDPeakFitRamda2(RTMessage Msg)
{
	if(Msg.LP.Hi != EN_UPDATE) return;
	float val = pTRamda2->GetValue();
	if(!pTRamda2->isValid) return;
	if(val <= 0.0) val = 1.5405;
	pPeak->ramda2 = val;
	UpdateGaussList();
}

void TPeakFitDialog::IDIRatio(RTMessage Msg)
{
	if(Msg.LP.Hi != EN_UPDATE) return;
	float val = pTIRatio->GetValue();
	if(!pTIRatio->isValid) return;
	pPeak->IRatio = val;
	UpdateGaussList();
}

void TPeakFitDialog::IDQRange(RTMessage Msg)
{
	if(Msg.LP.Hi != EN_UPDATE) return;
	float val = pTQRange->GetValue();
	if(!pTQRange->isValid) return;
	if(val < 1.0) val = 1.0;
	pPeak->QRange = val;
	UpdateGaussList();
}

void TPeakFitDialog::IDFWHMRatio(RTMessage Msg)
{
	if(Msg.LP.Hi != EN_UPDATE) return;
	if(pPeak->nPeak < 1) return;
	float val = pTfwhmRatio->GetValue();
	if(!pTfwhmRatio->isValid) return;
	if(val < 0.1) val = 0.1;
	pPeak->L_GfwhmRatio[pPeak->curPeak] = val;
	pTfwhmRatioScroll->SetPosition(5000);
}

void TPeakFitDialog::IDFWHMRatioScroll(RTMessage)
{
	if(pPeak->nPeak < 1) return;
	int pos = pTfwhmRatioScroll->GetPosition();
	float max = pPeak->L_GfwhmRatio[pPeak->curPeak];
	float val = max * (float)pos / 5000.0;
	if(val < 0.1) val = 0.1;
	pPeak->L_GfwhmRatio[pPeak->curPeak] = val;
	pTfwhmRatio->SetValue(val);
	if(val < 0.1) val = 0.1;
	val = pPeak->Gfwhm[pPeak->curPeak] / val;
	pPeak->Lfwhm[pPeak->curPeak] = val;
	pTLfwhm->SetValue(val);
}

void TPeakFitDialog::IDAngle(RTMessage Msg)
{
	if(Msg.LP.Hi != EN_UPDATE) return;
	if(pPeak->nPeak < 1) return;
	float val = pTAngle->GetValue();
	if(!pTAngle->isValid) return;
	if(val < 0.0) val = 0.0;
	pPeak->Angle[pPeak->curPeak] = val;
	pTAngleScroll->SetPosition(5000);
	UpdateGaussList();
}

void TPeakFitDialog::IDAngleScroll(RTMessage)
{
	if(pPeak->nPeak < 1) return;
	int pos = pTAngleScroll->GetPosition();
	float old = pPeak->Angle[pPeak->curPeak];
	float val = old - 5.0 + (float)pos/1000.0;
	pPeak->Angle[pPeak->curPeak] = val;
	pTAngle->SetValue(val);
}

void TPeakFitDialog::IDGaussFrac(RTMessage Msg)
{
	if(Msg.LP.Hi != EN_UPDATE) return;
	if(pPeak->nPeak < 1) return;
	float val = pTGFrac->GetValue();
	if(!pTGFrac->isValid) return;
	if(val < 0.0) val = 0.0;
	if(val > 1.0) val = 1.0;
	pPeak->GFrac[pPeak->curPeak] = val;
	int pos = (int)(val * 10000.0);
	pTGFracScroll->SetPosition(pos);
	UpdateGaussList();
}

void TPeakFitDialog::IDGFracScroll(RTMessage)
{
	if(pPeak->nPeak < 1) return;
	int pos = pTGFracScroll->GetPosition();
	pTGFrac->SetValue((float)pos/10000.0);
}

void TPeakFitDialog::IDHeight(RTMessage Msg)
{
	if(Msg.LP.Hi != EN_UPDATE) return;
	if(pPeak->nPeak < 1) return;
	float val = pTHeight->GetValue();
	if(!pTHeight->isValid) return;
	if(val < 0.0) val = 0.0;
	pPeak->Height[pPeak->curPeak] = val;
	pTHeightScroll->SetPosition(5000);
	UpdateGaussList();
}

void TPeakFitDialog::IDGaussfwhm(RTMessage Msg)
{
	if(Msg.LP.Hi != EN_UPDATE) return;
	if(pPeak->nPeak < 1) return;
	float val = pTGfwhm->GetValue();
	if(!pTGfwhm->isValid) return;
	if(val < 0.01) val = 0.01;
	pPeak->Gfwhm[pPeak->curPeak] = val;
	pTGfwhmScroll->SetPosition(5000);

	float r = pPeak->L_GfwhmRatio[pPeak->curPeak];
	if(r < 0.1) r = 0.1;
	val = val / r;
	pPeak->Lfwhm[pPeak->curPeak] = val;
	pTLfwhm->SetValue(val);
	UpdateGaussList();
}

void TPeakFitDialog::IDLorentzfwhm(RTMessage Msg)
{
	if(Msg.LP.Hi != EN_UPDATE) return;
	if(pPeak->nPeak < 1) return;
	float val = pTLfwhm->GetValue();
	if(!pTLfwhm->isValid) return;
	if(val < 0.01) val = 0.01;
	pPeak->Lfwhm[pPeak->curPeak] = val;
	pTLfwhmScroll->SetPosition(5000);

	float r = pPeak->L_GfwhmRatio[pPeak->curPeak];
	val = val * r;
	pPeak->Gfwhm[pPeak->curPeak] = val;

	UpdateGaussList();
}

void TPeakFitDialog::IDHeightScroll(RTMessage)
{
	if(pPeak->nPeak < 1) return;
	int pos = pTHeightScroll->GetPosition();
	float vmax = pPeak->Height[pPeak->curPeak];
	if(vmax == 0.0) vmax = 10000.0;
	pTHeight->SetValue((float)(vmax * (float)pos/5000.0));
}

void TPeakFitDialog::IDGaussfwhmScroll(RTMessage)
{
	if(pPeak->nPeak < 1) return;
	int pos = pTGfwhmScroll->GetPosition();
	float vmax = pPeak->Gfwhm[pPeak->curPeak];
	if(vmax == 0.0) vmax = 1.0;
	pTGfwhm->SetValue((float)(vmax * (float)pos/5000.0));
}

void TPeakFitDialog::IDLorentzfwhmScroll(RTMessage)
{
	if(pPeak->nPeak < 1) return;
	int pos = pTLfwhmScroll->GetPosition();
	float vmax = pPeak->Lfwhm[pPeak->curPeak];
	if(vmax == 0.0) vmax = 1.0;
	pTLfwhm->SetValue((float)(vmax * (float)pos/5000.0));
}

void TPeakFitDialog::IDPeakFitAnisotropic(RTMessage Msg)
{
	if(Msg.LP.Hi != EN_UPDATE) return;
	if(pPeak->nPeak < 1) return;
	float val = pTAnisotropic->GetValue();
	if(!pTAnisotropic->isValid) return;
	if(val < 0.0) val = 0.0;
	pPeak->Aniso[pPeak->curPeak] = val;
	int pos;
	if(val == 0) pos = 0;
	else pos = 5000;
	pTAnisotropicScroll->SetPosition(pos);
	UpdateGaussList();
}

void TPeakFitDialog::IDPeakFitAnisotropicScroll(RTMessage)
{
	if(pPeak->nPeak < 1) return;
	int pos = pTAnisotropicScroll->GetPosition();
	float vmax = pPeak->Aniso[pPeak->curPeak];
	if(vmax == 0.0) vmax = 0.02;
	pTAnisotropic->SetValue((float)(vmax * (float)pos/5000.0));
}

void TPeakFitDialog::IDBackA0(RTMessage Msg)
{
	if(Msg.LP.Hi != EN_UPDATE) return;
	float val = pTBackA0->GetValue();
	if(val < 0.0) val = 0.0;
	pPeak->Back[0] = val;
	pTBackA0Scroll->SetPosition(5000);
	UpdateGaussList();
}

void TPeakFitDialog::IDBackA0Scroll(RTMessage)
{
	int pos = pTBackA0Scroll->GetPosition();
	float max = pTBackA0->GetValue();
	if(max <= 0.0) max = 100.0;
	pTBackA0->SetValue(max * (float)pos/5000.0);
}

void TPeakFitDialog::IDBackA1(RTMessage Msg)
{
	if(Msg.LP.Hi != EN_UPDATE) return;
	float val = pTBackA1->GetValue();
	pPeak->Back[1] = val;
	pTBackA1Scroll->SetPosition(5000);
	UpdateGaussList();
}

void TPeakFitDialog::IDBackA1Scroll(RTMessage)
{
	int pos = pTBackA1Scroll->GetPosition();
	float max = pTBackA1->GetValue();
	if(max == 0.0) max = 100.0;
	pTBackA1->SetValue(max * (float)pos/5000.0);
}

void TPeakFitDialog::IDBackA2(RTMessage Msg)
{
	if(Msg.LP.Hi != EN_UPDATE) return;
	float val = pTBackA2->GetValue();
	pPeak->Back[2] = val;
	pTBackA2Scroll->SetPosition(5000);
	UpdateGaussList();
}

void TPeakFitDialog::IDBackA2Scroll(RTMessage)
{
	int pos = pTBackA2Scroll->GetPosition();
	float max = pTBackA2->GetValue();
	if(max == 0.0) max = 100.0;
	pTBackA2->SetValue(max * (float)pos/5000.0);
}

void TPeakFitDialog::IDAdd(RTMessage)
{
	int n = pPeak->nPeak;

	if(n+1 > MAX_FITPEAK)
	{
		MessageBox(HWindow, "nPeak exceeds MAX_FITPEAK (TPeakFitDialog::IDAdd)",
				"Warning!!", MB_OK);
		return;
	}
	if(n < 1)
	{
		n = 0;
		pPeak->Angle[n] = 30.0;
		pPeak->Height[n] = 1000.0;
		pPeak->GFrac[n] = 0.5;
		pPeak->Gfwhm[n] = 0.16;
		pPeak->Lfwhm[n] = 0.16;
		pPeak->L_GfwhmRatio[n] = 1.0;
		pPeak->Aniso[n] = 0.0;
	}
	else
	{
		pPeak->Angle[n] = pPeak->Angle[pPeak->curPeak];
		pPeak->Height[n] = pPeak->Height[pPeak->curPeak];
		pPeak->GFrac[n] = pPeak->GFrac[pPeak->curPeak];
		pPeak->Gfwhm[n] = pPeak->Gfwhm[pPeak->curPeak];
		pPeak->Lfwhm[n] = pPeak->Lfwhm[pPeak->curPeak];
		pPeak->L_GfwhmRatio[n] = pPeak->L_GfwhmRatio[pPeak->curPeak];
	}
	(pPeak->nPeak)++;
	pPeak->curPeak = pPeak->nPeak - 1;
	UpdateDialog();
//	InvalidateRect(Parent->HWindow, NULL, TRUE);
}

void TPeakFitDialog::IDDelete(RTMessage)
{
	int i;

	if(pPeak->curPeak < 0 || pPeak->curPeak >= pPeak->nPeak) return;
	if(pPeak->nPeak <= 1)
	{
		pPeak->nPeak = 0;
		UpdateDialog();
//		UpdateGaussList();
		return;
	}
	else
	{
		for(i = pPeak->curPeak ; i < pPeak->nPeak-1 ; i++)
		{
			pPeak->Angle[i] = pPeak->Angle[i+1];
			pPeak->Height[i] = pPeak->Height[i+1];
			pPeak->GFrac[i] = pPeak->GFrac[i+1];
			pPeak->Gfwhm[i] = pPeak->Gfwhm[i+1];
			pPeak->Lfwhm[i] = pPeak->Lfwhm[i+1];
			pPeak->L_GfwhmRatio[i] = pPeak->L_GfwhmRatio[i+1];
			pPeak->Aniso[i] = pPeak->Aniso[i+1];
		}
	}
	(pPeak->nPeak)--;
	if(pPeak->curPeak >= pPeak->nPeak) pPeak->curPeak = pPeak->nPeak-1;
	UpdateDialog();
//	InvalidateRect(Parent->HWindow, NULL, TRUE);
//	UpdateGaussList();
}

void TPeakFitDialog::Ok(RTMessage Msg)
{
	if(pTAutoRedraw->GetCheck() == BF_CHECKED) pPeak->isAutoRedraw = TRUE;
	else pPeak->isAutoRedraw = FALSE;
	TDialog::Ok(Msg);
}

void TPeakFitDialog::Cancel(RTMessage Msg)
{
	memcpy(&back, pPeak, sizeof(PeakFitData));
	TDialog::Cancel(Msg);
}

void TPeakFitDialog::IDView(RTMessage)
{
	InvalidateRect(Parent->HWindow, NULL, TRUE);
}

