// PeakSearch.h
//
//////////////////////////////////////////////////////////////////////

#if !defined(AFX_PEAKSEARCH_H__ECB1C19C_9580_11D1_84E9_485223000000__INCLUDED_)
#define AFX_PEAKSEARCH_H__ECB1C19C_9580_11D1_84E9_485223000000__INCLUDED_

#if _MSC_VER >= 1000
#pragma once
#endif // _MSC_VER >= 1000

#include "RealDC.h"

static const int PEAKSEARCH_MAXPEAKS = 100;

static const int PEAKSEARCH_SPECTRUMKIND_INVALID = 0;
static const int PEAKSEARCH_SPECTRUMKIND_XRD     = 1;
static const int PEAKSEARCH_SPECTRUMKIND_RAMAN   = 2;
static const int PEAKSEARCH_SPECTRUMKIND_IR      = 2;

#include "..\VariousData.h"
class CVariousData;

class CPeakSearch
{
public:
	int SetYAxisType(int type);
	bool IsPeakFitPFCFile(LPCSTR fname);
	bool IsUsePeakFit(LPCSTR LSQProgram);
	double m_PercentSigma;
	double m_Sigma;
	int m_iPrt, m_iPrtIntvl;
	int m_BGOrder;
	int m_SpectrumKind;
	int m_iCurrentPeak;
	int CalcPeak(int iCurve);
	int CalcPeakFitPeak(int iCurve);
	int CalcRmnFitPeak(int iCurve, bool IsUsePeakFitBG);
	CVariousData *m_pOriginalData;
	CGraphData m_BG, m_Total;
	CGraphDataArray m_Curves;
	int m_MaxIter;
	CString m_LSQMethod;
	CString m_DataPath;
	CString m_XRDLSQProgram;
	CString m_RamanLSQProgram;
	CString m_Title;
	double m_DumpingFactor;
	double m_QRange;
	double m_ConvergenceCriteria;
	double m_LKa1, m_LKa2, m_IRatio;
	double m_OptRegionX0, m_OptRegionX1;
	double m_OutputRegionX0, m_OutputRegionX1;
	int m_MaxPeaks;
	double m_BGb[5];
	int    m_OptimizeBGb[5];
	double m_DiffBGb[5];
	int m_nPeaks;
	double *m_Position;
	int    *m_OptimizePosition;
	double *m_DiffPosition;
	double *m_CTotal;
	int    *m_OptimizeCTotal;
	double *m_DiffCTotal;
	double *m_GaussFraction;
	int    *m_OptimizeGaussFraction;
	double *m_DiffGaussFraction;
	double *m_CGauss;
	int    *m_OptimizeCGauss;
	double *m_DiffCGauss;
	double *m_FWHMGauss;
	int    *m_OptimizeFWHMGauss;
	double *m_DiffFWHMGauss;
	double *m_CLorentz;
	int    *m_OptimizeCLorentz;
	double *m_DiffCLorentz;
	double *m_FWHMLorentz;
	int    *m_OptimizeFWHMLorentz;
	double *m_DiffFWHMLorentz;
	double *m_FWHMGLRatio;
	int    *m_OptimizeFWHMGLRatio;
	double *m_DiffFWHMGLRatio;
	double *m_Aniso;
	int    *m_OptimizeAniso;
	double *m_DiffAniso;

	CPeakSearch();
	~CPeakSearch();
	int ReadPFCFile(LPCSTR fname, bool ReadSavedData = false);
	int ReadPeakFitPFCFile(LPCSTR fname, bool IsXRD, bool ReadSavedData = false);
	int ReadRmnFitPFCFile(LPCSTR fname, bool ReadSavedData = false);
	int SavePFCFile(LPCSTR savefname);
	int SavePeakFitPFCFile(LPCSTR savefname, LPCSTR LSQProgram);
	int SaveRmnFitPFCFile(LPCSTR savefname);
	int SetMaxPeaks(int MaxPeaks);
	int RemoveAll();
	int AddPeak(double position, double ctotal, double gaussfraction, 
								double fwhmglratio, double fwhm);
	int DeletePeak(int i);
};


#endif // !defined(AFX_PEAKSEARCH_H__ECB1C19C_9580_11D1_84E9_485223000000__INCLUDED_)
