package AtomType;
use Crystal::AtomTypeObject;
@ISA = qw(AtomTypeObject);

#JTu[`
@EXPORT = qw();

use strict;
use Deps;
use ProgVars;
use CSV;

my $DirectorySeparator = '\\';
my $LF = "\n";

#===============================================
# pXiǂݍDBj
#===============================================
my $WebRootDir      = ProgVars::WebRootDir(); #"d:\\MyWebs";
my $ProgramDir      = ProgVars::ProgramDir(); #"d:\\Programs";
my $BinDir          = ProgVars::BinDir();
my $DBDir           = ProgVars::PeriodicTableDir();

#my $DBDir = Deps::MakePath($WebRootDir, "Research", 0);
#   $DBDir = Deps::MakePath($DBDir, "Databases", 0);

my $NonrelInputPath           = Deps::MakePath($DBDir, "nonrel",                      0);
my $IonRadius1DBPath          = Deps::MakePath($DBDir, "IonRadius1.html",             0);
my $IonRadius2DBPath          = Deps::MakePath($DBDir, "IonRadius2.html",             0);
my $PeriodicTable1DBPath      = Deps::MakePath($DBDir, "PeriodicTable1.html",         0);
my $PeriodicTable2DBDir       = Deps::MakePath($DBDir, "PeriodicTable2",              0);
my $PeriodicTable3DB          = Deps::MakePath($DBDir, "PERIODIC.CSV",                0);
my $ElectronicConfigurationDB = Deps::MakePath($DBDir, "ElectronicConfiguration.csv", 0);

my $MXDDir     = ProgVars::MXDDir();
my $MXDExecDir = ProgVars::MXDExecDir();
my $SX1DBPath  = ProgVars::SX1DBPath();

sub GetAtomDBHash {
	my ($element, $charge, $nCoordination) = @_;
	$charge = 0 if(!defined $charge);

	if($element =~ /^\d+$/) {
		$element = AtomType::GetAtomName($element);
	}

	my ($mass, $charge0, $ai, $bi, $ci, $rad) = AtomType::GetBusingParameter($element);
	my ($AtmNum, $JName, $EName, $NameOrigin) = AtomType::GetIonInfFromPeriodicTable1($element);
	my ($AtomicNumber, $AtomName, $Charge)    = AtomType::GetAtomInformation($element);
	my $CationRadius = AtomType::GetCationRadius($element, $charge, $nCoordination);
	my @IonRadius    = AtomType::GetIonRadius2($element, $charge, $nCoordination);
	my ($filepath, $AtomJName, $AtomEName,
		$AtomicMass, $FoundYear, $AtomicRadius,
		$MeltingPoint, $BoilingPoint, $Density,
		$HeatCapacity, $IonizationPotential, $ElectronAffinity) = AtomType::GetIonInfFromPeriodicTable2($AtmNum);
	my $NameOrigin2;
	($AtmNum, $JName, $EName, $NameOrigin2) = AtomType::GetIonInfFromPeriodicTable1($element);
	my ($pOrbitalName1, $pnElectron1)         = AtomType::GetElectronicConfiguraton($AtmNum);
	my ($pOrbitalName2, $pnElectron2)         = AtomType::SpeculateElectronicConfiguration($AtmNum);
	my $ElectronConfiguration1 = '';
	my $ElectronConfiguration2 = '';
	for(my $i = 0 ; $i < @$pOrbitalName1 ; $i++) {
		next if($pnElectron1->[$i] == 0);
		$ElectronConfiguration1 .= "$pOrbitalName1->[$i]($pnElectron1->[$i])";
	}
	for(my $i = 0 ; $i < @$pOrbitalName2 ; $i++) {
		next if($pnElectron2->[$i] == 0);
		$ElectronConfiguration2 .= "$pOrbitalName2->[$i]($pnElectron2->[$i])";
	}
#print "  $ElectronConfiguration1\n";
#print "  $ElectronConfiguration2\n";

	$AtomicMass = $mass        if(!defined $AtomicMass);
	$JName      = $AtomJName   if(!defined $JName);
	$EName      = $AtomEName   if(!defined $EName);
	$NameOrigin = $NameOrigin2 if(!defined $NameOrigin);
	my %db = (
		name                   => $element,
		Z                      => $AtmNum,
		mass                   => $AtomicMass,
		charge                 => $charge,
		nCoordination          => $nCoordination,
		AtomicRdius            => $AtomicRadius,
		CationRadius           => $CationRadius,
		pIonRadius             => \@IonRadius,
		IonizationPotential    => $IonizationPotential,
		ElectronAffinity       => $ElectronAffinity,
		MeltingPoint           => $MeltingPoint,
		BoilingPoint           => $BoilingPoint,
		Density                => $Density,
		HeatCapacity           => $HeatCapacity,
		ElectronConfiguration1 => $ElectronConfiguration1,
		ElectronConfiguration2 => $ElectronConfiguration2,
		ai                     => $ai,
		bi                     => $bi,
		ci                     => $ci,
		rad                    => $rad,
		JName                  => $JName,
		EName                  => $EName,
		NameOrigin             => $NameOrigin,
		FoundYear              => $FoundYear,
		
		);

	return \%db;
}

sub GetShellName
{
	my ($n) = @_;
	my @names = qw(a K L M N O P);
	return $names[$n];
}

sub GetOrbName
{
	my ($n, $l, $m) = @_;
	if(!defined $m) {
		my @names = ('s', 'p', 'd', 'f', 'g', 'h');
		return "$n$names[$l]";
	}

	my @names1 = ('s'); 										# for l=0
	my @names2 = ('py', 'pz', 'px');							# for l=1, m=-1, 0, 1
	my @names3 = ('dxy', 'dyz', 'd3z2-r2', 'dzx', 'dx2-y2');	# for l=2, m=-2,-1,0,1,2
	my @names4 = ('fy(3x2-y2)', 'fxyz', 'fy(5z2-r2)', 'fz(5z2-3r2)',
		      'fx(5z2-r2)', 'fz(x2-y2)', 'fx(x2-3y2)');		# for l=3, m=-3,-2,-1,0,1,2,3
	my @names = (\@names1, \@names2, \@names3, \@names4);

	my $pNames = $names[$l];
	my $name = $pNames->[$m+$l];
	return "$n$name";
}


sub GetEnglishName
{
	my ($name) = (@_);
	my ($AtmNum, $JName, $EName, $NameOrigin)
		= GetIonInfFromPeriodicTable1($name);
	return $EName;
}

sub GetBusingParameter
{
	my ($atomname, $Debug) = @_;

	$atomname =~ s/\{.*?\}//g;
	$atomname =~ s/\d//g;
#print "an: $atomname\n";
#print "DB: $SX1DBPath\n";
	unless(open(IN, "<$SX1DBPath")) {
		print("Error in MXD.pm::GetBusingParameter: "
			."Can not read DB [$SX1DBPath].<br>\n") if($Debug);
		return ('');
	}
	<IN>;
	while(<IN>) {
		my ($name, $mass, $charge, $ai, $bi, $ci, $rad) = split(/\s+/, $_);
#print "uc: $name eq $atomname\n";
		if(uc $name eq uc $atomname) {
			close(IN);
			return ($mass, $charge, $ai, $bi, $ci, $rad);
		}
	}
	
	close(IN);
	return ('');
}

sub FillAtomTypeData
{
	my ($this) = @_;
	my $name = $this->AtomNameOnly();
	my ($mass, $charge, $ai, $bi, $ci, $rad) = AtomType::GetBusingParameter($name);
$name =~ s/\{.*?\}//g;
$name =~ s/\d//g;
#print "an: $name: $mass\n";

	$this->SetCharge($charge) if($this->Charge() == 0.0);
	$this->SetAtomicMass($mass) if($this->AtomicMass() <= 0.0);
#print "<H1>name: $name  charge: $charge<br></H1>\n";
}

sub GetAtomInformation
{
	my ($name) = (@_);

	my ($AtomName, $AtomicNumber, $Charge);
	chomp($name);
	$name =~ s/^\s*//;
	$name = uc $name;

	if($name =~ /^\d+$/) {
		$AtomicNumber = $name;
		$AtomName = &GetAtomName($AtomicNumber);
		$Charge = 0;
	}
	else {
		my ($name, $charge) = ($name =~ /^(.*?)\s*([\d\+-]+)?$/);
#print "name=$name  charge=$charge\n";
		if(!defined $name) {
			$AtomName = '';
			$Charge = 0;
		}
		elsif(!defined $charge) {
			$AtomName = $name;
			$Charge = 0;
		}
		else {
			$AtomName = $name;
			$Charge   = &ChargeToVal($charge);
#			$Charge   =  1.0 if($Charge eq '+');
#			$Charge   = -1.0 if($Charge eq '-');
#print "  c=$Charge\n";
		}
		$AtomicNumber = &GetAtomicNumber($AtomName);
	}
	$AtomName = ucfirst lc $AtomName;
	return ($AtomicNumber, $AtomName, $Charge);
}

sub GetAtomName
{
	my ($AtomNumber, $Debug) = (@_);

	my $name = '';
	unless(open(IN, "<$NonrelInputPath")) {
		print "Error: $!: $NonrelInputPath$LF$LF";# if($Debug);
		return -1;
	}
# H  HFS ATOM CALC.
#    1  300    1
#   1.00000   0.97804  20.00000  32.00000   0.00000
#  1.0  0.0  0.0   -0.30       1.0
	while(<IN>) {
		my $line = $_;       # H  HFS ATOM CALC.
		my $nextline = <IN>; #    1  300    1
		$nextline = <IN>;    #   1.00000   0.97804  20.00000  32.00000   0.00000
		my ($n) = ($nextline =~ /^\s*([\d\.]+?)\s/);
		if($n == $AtomNumber) {
			($name) = ($line =~ /^\s*(\w+?)\s/);
			last;
		}
# ̋s܂ŃXLbv
		while(<IN>) {
			my $line2 = $_;
			chomp($line2);
			$line2 =~ s/^\s*//;
			if($line2 eq '') {
				last;
			}
		}
	}
	close(IN);
	$name = ucfirst lc $name;
	return $name;
}

sub GetAtomicNumber 
{
	my ($name, $Debug) = (@_);

	$name =~ s/\{.*?\}//g;
	$name =~ s/\d//g;
	$name = uc $name;
	$name =~ s/^(\D+?)\d.*$/$1/;
	my $an = -1;
	unless(open(IN, "<$NonrelInputPath")) {
		print "Error: $!: $NonrelInputPath${LF}";# if($Debug);
		return -1;
	}

	while(<IN>) {
# H  HFS ATOM CALC.
#    1  300    1
#   1.00000   0.97804  20.00000  32.00000   0.00000
#  1.0  0.0  0.0   -0.30       1.0
  		if($_ =~ /^\s*$name\s/) { # H  HFS ATOM CALC.
			my $line = <IN>;  #    1  300    1
			$line = <IN>;     #   1.00000   0.97804  20.00000
			($an) = ($line =~ /^\s*(.*?)\s/);
			last;
		}
	}
	close(IN);
	return $an;
}

sub ChargeToVal
{
	my ($charge) = @_;

	return  0 if(!defined $charge);
	return  1 if($charge eq '+');
	return -1 if($charge eq '-');

#print("charge=$charge");
	if($charge =~ /(\d+)\+/ or $charge =~ /\+(\d+)/) {
#print(" +");
		return + abs($1);
	}
	if($charge =~ /(\d+)\-/ or $charge =~ /\-(\d+)/) {
#print(" -");
		return - abs($1);
	}
	return $charge;
}

sub ChargeToStr
{
	my ($charge) = @_;

	return ''   if(!defined $charge);
	return '1+' if($charge eq '+');
	return '1-' if($charge eq '-');
	if($charge =~ /(\d+)\+/ or $charge =~ /\+(\d+)/) {
		$charge = + abs($1);
	}
	if($charge =~ /(\d+)\-/ or $charge =~ /\-(\d+)/) {
		$charge = - abs($1);
	}
	$charge = '' if($charge == 0.0);
	return $charge;
}

sub GetCationRadius
{
	my ($name, $charge, $nCoordination, $Debug) = (@_);
	$name   = uc $name;

	$charge = &ChargeToVal($charge);

	my $r = -1.0;
	unless(open(IN, "<$IonRadius1DBPath")) {
		print "Error: $!: $IonRadius1DBPath${LF}";# if($Debug);
		return -1;
	}
	while(<IN>) {
		my $line = $_;
		next if($line !~ m|<TR>|i);
		my $atmline    = <IN>;
		my $atmnumline = <IN>;
		my $radiusline = <IN>;
		$atmline =~ m|<TD.*?>(.*?)<SUP>(.*?)<\/SUP><\/TD>|i;
		my $atname   = (defined $1)? uc $1 : '';
		my $atcharge = 0;
		if(defined $2) {
			$atcharge = $2;
		}
#		$atcharge = 0.0 + $atcharge if($atcharge =~ /\+/);
		$atcharge =~ s/\+//;
		if($atcharge =~ /\-/) {
			$atcharge =~ s/-//;
			$atcharge = 0.0 - $atcharge;
		}
		$atmnumline =~ m|<TD.*?>(.*?)<\/TD>|i;
		my $atomicnumber   = $1;
		$radiusline =~ m|<TD.*?>(.*?)<\/TD>|i;
		my $radius   = $1;

		if($name eq $atname and $charge eq $atcharge) {
			close(IN);
			return $radius;
		}
		$r = $radius if($name eq $atname);
	}
	close(IN);
	return $r;
}

sub GetIonRadius2
{
	my ($name, $charge, $nCoord, $Debug) = (@_);

	$name = uc $name;
	my @radius;
	
	unless(open(IN, "<$IonRadius2DBPath")) {
		print "Error: $!: $IonRadius2DBPath${LF}";# if($Debug);
		return -1;
	}
	while(<IN>) {
		my $line = $_;
		$line =~ m|\<P\>\<B\>(.*?)\</B\>\</P\>|i;
		my $AtomName = $1;
		next unless(defined $AtomName and uc $AtomName eq $name);

		my $rline = '';
		while(<IN>) {
			my $line2 = $_;
			last if($line2 =~ m|\<TD\s+align=middle.*?\>|i);
			chomp($line2);
			$rline .= $line2;
		}
		my @a = split(/(\<\/?B\>|\<\/?FONT.*?\>|\<BR\>|\<\/?P\>|\<I\>|\<P.*?\>)/i, $rline);;
		my $IsPauling = 0;
		my $IsReliable = 0;
		foreach my $s (@a) {
			$IsPauling = 1 if($s =~ /\<I\>/i);
			$IsReliable = 1 if($s =~ /\<B\>/i);
			$s =~ s/^\s+//;
			$s =~ s/\s+$//;
			next if($s eq '' or $s =~ /[\<\>]/);
			if($s =~ /^[\d\.]+$/) { # to A
				$s = $s * 0.01;
				$s = "$s (less reliable)"
					unless($IsReliable);
				$s = "$s (deduced by Pauling(1960) or Ahrens(1952))"
					if($IsPauling);
			}
			push(@radius, $s);
			$IsPauling = 0;
			$IsReliable = 0;
		}
		last;
	}
	close(IN);
	return @radius;
}

sub GetAtomRadius
{
	my ($atnum, $UseApproximation, $Debug) = (@_);
	my ($filepath, $AtomJName, $AtomEName,
		$AtomicMass, $FoundYear, $AtomicRadius,
		$MeltingPoint, $BoilingPoint, $Density,
		$HeatCapacity, $IonizationPotential, $ElectronAffinity) = &GetIonInfFromPeriodicTable2($atnum, $Debug);
	return $AtomicRadius;
}

sub GetIonInfFromPeriodicTable2
{
	my ($atnum, $Debug) = (@_);
	my $AtomJName  = '';
	my $AtomEName  = '';
	my $AtomicMass = '';
	my $FoundYear  = '';
	my $AtomicRadius = '';
	my $MeltingPoint = '';
	my $BoilingPoint = '';
	my $Density      = '';
	my $HeatCapacity = '';
	my $IonizationPotential = '';
	my $ElectronAffinity    = '';
#qԍ 94 
#q 239 
#N (N) 1940 
#qa (A) 1.64 
#Z_ (oC) 639.5 
#_ (oC) 3235 
#x (g/cm3) @ 
#M (cal/g oC) @ 
#CIGlM[ (eV) 5.8 
#dqea (eV) 

	my $fname = sprintf "A%03dj.html", int($atnum+0.1);
	my $filepath = Deps::MakePath($PeriodicTable2DBDir, $fname, 0);
	unless(open(IN, "<$filepath")) {
		print "Error: $!: $filepath${LF}";# if($Debug);
		return -1;
	}
	while(<IN>) {
		my $line = $_;
#print "line: $line\n";
		if($line =~ m|<title>(.*?)<\/title>|i) {
			$AtomEName = $1;
		}
		elsif($line =~ m|\<font SIZE=\"\+3\">(.*?)(\(.*?\))?\<\/font\>|i) {
			$AtomJName = $1;
		}
		elsif($line =~ m|qԍ|i) {
			$line =~ /\>(\d+)\<\/font/;
		}
		elsif($line =~ m|q|i) {
			$line =~ /1\">([\d\.]*)<\/font/;
			$AtomicMass = $1;
		}
		elsif($line =~ m|N.*\>(\d*?)\<\/font|i) {
			$FoundYear = $1;
		}
		elsif($line =~ m|qa.*\>([\d\.]*).*?\<\/font|i) {
			$AtomicRadius = $1;
		}
		elsif($line =~ m|Z_.*\>([-\+\d\.]*)\<\/font|i) {
			$MeltingPoint = $1;
		}
		elsif($line =~ m|_.*\>([-\+\d\.]*)\<\/font|i) {
			$BoilingPoint = $1;
		}
		elsif($line =~ m|x.*\>([\d\.]*)\<\/font|i) {
			$Density = $1;
		}
		elsif($line =~ m|M.*\>([\d\.]*)\<\/font|i) {
			$HeatCapacity = $1;
		}
		elsif($line =~ m|CI.*\>([-\+\d\.]*)\<\/font|i) {
			$IonizationPotential = $1;
		}
		elsif($line =~ m|dq.*\>([-\+\d\.]*)\<\/font|i) {
			$ElectronAffinity = $1;
		}
	}
	close(IN);
	return ($filepath, $AtomJName, $AtomEName,
		$AtomicMass, $FoundYear, $AtomicRadius,
		$MeltingPoint, $BoilingPoint, $Density,
		$HeatCapacity, $IonizationPotential, $ElectronAffinity);
}

sub GetIonInfFromPeriodicTable1
{
	my ($name, $Debug) = (@_);

	$name = uc $name;
	my $AtmNum = -1;
	my $JName = "";
	my $EName = "";
	my $NameOrigin = "";
#    <TD>ԍ</TD>
#    <TD>L</TD>
#    <TD>f</TD>
#    <TD>pꖼ</TD>
#    <TD>O̗R</TD></TR>

#print "=== Open [$PeriodicTable1DBPath]\n";
	unless(open(IN, "<$PeriodicTable1DBPath")) {
		print "Error: $!: $PeriodicTable1DBPath${LF}";# if($Debug);
		return -1;
	}

	while(<IN>) {
		my $line = $_;
		next if($line !~ m|<TD><A name=.*?>(\d+)<\/A><\/TD>|i);
		$AtmNum = $1;
		$line    = <IN>;
		$line =~ m|<TD.*?>(.*?)<\/TD>|i;
		my $atname   = $1;
#print "at: $AtmNum: [$atname] [$name]\n";
#		next if($name ne uc $atname and ($name =~ /^[\d+-]+$/ and $AtmNum != int($name+0.1)) ) ;
		next if($name ne uc $atname) ;
		next if($name =~ /^[\d\+\-]+$/ and $AtmNum != int($name+0.1));

		$line    = <IN>;
		$line =~ m|<TD.*?>(.*?)<\/TD>|i;
		$JName = $1;
		$line    = <IN>;
		$line =~ m|<TD.*?>(.*?)<\/TD>|i;
		$EName = $1;
		$line    = <IN>;
		$line =~ m|<TD.*?>(.*?)<\/TD>|i;
		$NameOrigin = $1;
		last;
	}
	close(IN);

	return ($AtmNum, $JName, $EName, $NameOrigin);
}

sub GetAtomInformationByKey
{
	my ($atomname, $key, $Debug) = @_;

	my ($AtomicNumber, $IonName, $IonCharge) = &GetAtomInformation($atomname);

	unless(open(IN, "<$PeriodicTable3DB")) {
		print "Can't open $PeriodicTable3DB$LF" if($Debug);
		print "$LF" if($Debug);
		return;
	}
	my $line = <IN>;
	my @DBTitle = split(/,/, $line);
	my @DBValue;
	while(<IN>) {
		@DBValue = split(/,/, $_);
		last if($DBValue[0] == $AtomicNumber);
	}
	close(IN);

	my $i;
	for($i = 0 ; $i < @DBTitle ; $i++) {
		last if($DBTitle[$i] =~ /$key/);
	}
	return $DBValue[$i];
}

sub GetElectronicConfiguraton
{
	my ($nElectron) = @_;
	
	my $CSV = new CSV;
	if(!$CSV->Read($ElectronicConfigurationDB, 0)) {
		return undef;
	}

	my $pLabelArray   = $CSV->LabelArray();
	my $pDataArray    = $CSV->DataArray();
	my $pAtomNumArray = $pDataArray->[2];
	my $nAtomNumArray = @$pAtomNumArray;
	my $i = 0;
	for( ; $i < $nAtomNumArray ; $i++) {
#print "i=$i: $pAtomNumArray->[$i] / $nElectron\n";
		last if($nElectron == $pAtomNumArray->[$i]);
	}
	return undef if($i == $nAtomNumArray);

#print "ne=$nElectron: i=$i\n";
	my (@name, @n);
	for(my $j = 0 ; ; $j++) {
		my $p = $pDataArray->[3+$j];
		last if(!defined $p);
		my $ne = $p->[$i];
		last if(!defined $ne);
		$ne = 0 if($ne eq '');
		$name[$j] = $pLabelArray->[3+$j];
		$n[$j]    = $ne;
#print "$j: $name[$j]: $n[$j]\n";
	}
	return (\@name, \@n);
}

sub SpeculateElectronicConfiguration
{
	my ($nElectron) = @_;

	my $ResCharge = $nElectron;

	my $n1s = 0.0;
	if($ResCharge > 2.0) {
		$n1s = 2.0;
		$ResCharge -= 2.0;
	}
	else {
		$n1s = $ResCharge;
		$ResCharge = 0.0;
	}
	my $n2s = 0.0;
	if($ResCharge > 2.0) {
		$n2s = 2.0;
		$ResCharge -= 2.0;
	}
	else {
		$n2s = $ResCharge;
		$ResCharge = 0.0;
	}
	my $n2p = 0.0;
	if($ResCharge > 6.0) {
		$n2p = 6.0;
		$ResCharge -= 6.0;
	}
	else {
		$n2p = $ResCharge;
		$ResCharge = 0.0;
	}
	my $n3s = 0.0;
	if($ResCharge > 2.0) {
		$n3s = 2.0;
		$ResCharge -= 2.0;
	}
	else {
		$n3s = $ResCharge;
		$ResCharge = 0.0;
	}
	my $n3p = 0.0;
	if($ResCharge > 6.0) {
		$n3p = 6.0;
		$ResCharge -= 6.0;
	}
	else {
		$n3p = $ResCharge;
		$ResCharge = 0.0;
	}
	my $n3d = 0.0;
	my $n4s = 0.0;

	if($ResCharge > 2.0) {
		$n4s = 2.0;
		$ResCharge -= 2.0;
	}
	else {
		$n4s = $ResCharge;
		$ResCharge = 0.0;
	}
	if($ResCharge > 10.0) {
		$n3d = 10.0;
		$ResCharge -= 10.0;
	}
	else {
		$n3d = $ResCharge;
		$ResCharge = 0.0;
	}

	my $n4p = 0.0;
	if($ResCharge > 6.0) {
		$n4p = 6.0;
		$ResCharge -= 6.0;
	}
	else {
		$n4p = $ResCharge;
		$ResCharge = 0.0;
	}
	my $n4d = $ResCharge;

	my @name = qw(1s    2s    2p    3s    3p    3d    4s    4p    4d);
	my @ne   =   ($n1s, $n2s, $n2p, $n3s, $n3p, $n3d, $n4s, $n4p, $n4d);
	return (\@name, \@ne);
}

BEGIN {
}

sub new
{
	my ($module) = @_;
	my $this = {};
	bless $this;
	return $this;
}

sub DESTROY
{
	my $this = shift;
}

sub SetAtomicNumber
{
	my ($this, $AtomicNumber);
	AtomTypeObject::SetAtomicNumber($this, $AtomicNumber);

	&SetAtomInformation();

	return $this->{'AtomNumber'};
}

sub SetAtomName
{
	my ($this,$AtomName) = @_;
	AtomTypeObject::SetAtomName($this, $AtomName);
	my $AtomicNumber      = &GetAtomicNumber($AtomName);
	$this->{AtomicNumber} = $AtomicNumber;

	$this->SetAtomInformation();

	return $AtomName;
}

sub SetAtomInformation
{
	my ($this) = @_;

	my $AtomName = $this->AtomNameOnly(); #$this->{'AtomName'};
$AtomName =~ s/\{.*?\}//g;
$AtomName =~ s/\d//g;
	my $Charge       = $this->Charge(); #$this->{'Charge'};
	my $AtomicNumber = $this->AtomicNumber(); #$this->{'AtomicNumber'};
	if($AtomicNumber <= 0) {
		($AtomicNumber) = &GetIonInfFromPeriodicTable1($AtomName);
	}
#print "an: [$AtomName] #$AtomicNumber  Charge=$Charge<br>\n";
	my $IonRadius = &GetCationRadius($AtomName, $Charge, 6);
#	if($IonRadius <= 0) {
#		my @rad = &GetIonRadius2($AtomName, $Charge, 6);
#		$IonRadius = $rad[2];
#	}
	my ($filepath, $AtomJName, $AtomEName,
		$AtomicMass, $FoundYear, $AtomicRadius,
		$MeltingPoint, $BoilingPoint, $Density,
		$HeatCapacity, $IonizationPotential, $ElectronAffinity)
			= &GetIonInfFromPeriodicTable2($AtomicNumber);
#	my ($AtmNum, $JName, $EName, $NameOrigin)
#		= &GetIonInfFromPeriodicTable1($AtomicName);
#print "an: $AtomName: $AtomicMass $AtomicNumber\n";

	$this->{AtomicRadius}        = $AtomicRadius;
	$this->{IonRadius}           = $IonRadius;
	$this->{AtomicMass}          = $AtomicMass;
	$this->{IonizationPotential} = $IonizationPotential;
	$this->{ElectronAffinity}    = $ElectronAffinity;
	return 1;
}

1;