import numpy as np
from numpy import sin, cos, tan, arccos, arcsin, arctan, sqrt, exp, log, pi

from tklib.tkobject import tkObject
from tklib.tkutils import IsDir, IsFile, SplitFilePath
from tklib.tkutils import terminate, pint, pfloat, getarg, getintarg, getfloatarg
from tklib.tkutils import SplitFilePath, GetList
from tklib.tkre    import DelQuote, DelSpace
import tklib.tkre

# バンド構造をプロット
# xk: プロットするk点の蓄積距離のリスト
# yE: E(xk[i])。入れ子になったリストで構わない
# ktotallist: k点境界における、最初のk点からの距離の和のリスト
# ktotal_namelist: k点境界における、k点の名称
def plot_band(plt, axis, ISPIN, xk, yEup, yEdn, Erange, occups = None, occdns = None, 
              ktotallist = None, ktotal_namelist = None,
              nmax = 1.0, EV = None, EC = None, EF = None, EHOMO = None, ELUMO = None,
              EVlabel = '$E_V$', EClabel = '$E_C$', EFlabel = '$E_F$', 
              EHOMOlabel = '$E_{HOMO}$', ELUMOlabel = '$E_{LUMO}$',
              marker = 'o', linestyle = '', color = 'black', linewidth = 0.3,
              markersize = 2.0, markeredgewidth = 0.5,
              legendloc = 'best',
              fontsize = 16, labelsize = 12, legend_fontsize = 12):

# 表示範囲は決め打ち
    xlim = [min(xk), max(xk)]
    axis.set_xlim(xlim)
    axis.set_ylim(Erange)

#    print("tkbandstructre.plot_band: Plot EV etc")
    if EV is not None:
        axis.plot(xlim, [EV, EV], label = EVlabel, linestyle = 'dashed', linewidth = 0.5, color = 'red')
    if EC is not None:
        axis.plot(xlim, [EC, EC], label = EClabel, linestyle = 'dashed', linewidth = 0.5, color = 'red')
    if EF is not None:
        axis.plot(xlim, [EF, EF], label = EFlabel, linestyle = 'dashed', linewidth = 0.5, color = 'blue')
    if EHOMO is not None:
        axis.plot(xlim, [EHOMO, EHOMO], label = EHOMOlabel, linestyle = 'dashed', linewidth = 0.5, color = 'green')
    if ELUMO is not None:
        axis.plot(xlim, [ELUMO, ELUMO], label = ELUMOlabel, linestyle = 'dashed', linewidth = 0.5, color = 'green')
    
#    print("tkbandstructre.plot_band: Plot band structure")

# バンド構造をプロット
# 色の濃淡を表現するために16進数文字のリストを準備
    convchar = ['0', '1', '2', '3', '4', '5', '6', '7', '8', '9', 'a', 'b', 'c', 'd', 'e', 'f']
#    print("k point names:", ktotal_namelist)
    if occups is None:
        axis.plot(xk, yEup, linestyle = linestyle, color = color, linewidth = linewidth,
                marker = marker, markerfacecolor = 'none', markeredgecolor = 'black', 
                markeredgewidth = markeredgewidth, markersize = markersize)
        if ISPIN == 2:
            axis.plot(xk, yEdn, linestyle = linestyle, color = color, linewidth = linewidth,
                    marker = marker, markerfacecolor = 'none', markeredgecolor = 'red', 
                    markeredgewidth = markeredgewidth, markersize = markersize)
    else:
        for ik in range(len(xk)):
            for iE in range(len(yEup[ik])):
                ne = occups[ik][iE]
                if ne < 0.0:
                    ne = 0.0
                elif ne > nmax:
                    ne = nmax
                idx = int((nmax - ne) / nmax * 15)
                color = convchar[idx] * 6
                axis.plot(xk[ik], yEup[ik][iE], linestyle = 'none', color = color, linewidth = linewidth,
                        marker = marker, markerfacecolor = "#" + color, markeredgecolor = 'black', 
                        markeredgewidth = markeredgewidth, markersize = markersize)
                if ISPIN == 2:
                    ne = occdns[ik][iE]
                    if ne < 0.0:
                        ne = 0.0
                    elif ne > nmax:
                        ne = nmax
                    idx = int((nmax - ne) / nmax * 15)
                    color = 'ff' + convchar[idx]*4
                    axis.plot(xk[ik], yEdn[ik][iE], linestyle = 'none', color = color, linewidth = linewidth,
                            marker = marker, markerfacecolor = "#" + color, markeredgecolor = 'red', 
                            markeredgewidth = markeredgewidth, markersize = markersize)
    
    if ktotallist is not None:
# Γ点、BZ境界の縦線を引く
        for i in range(1, len(ktotallist)):
            axis.plot([ktotallist[i-1], ktotallist[i-1]], Erange, 
                        linestyle = '-', color = 'black', linewidth = 0.5)

# k軸の目盛りにk点の名称を表示する
# グラフ枠が一つであれば plt.xtics()で設定できる
# axisに対しては、.setpでattributeを直接書き換える必要があるらしい
        plt.setp(axis, xticks = ktotallist, xticklabels = ktotal_namelist)

    axis.set_xlabel("$k$", fontsize = fontsize)
    axis.set_ylabel("$E$ (eV)", fontsize = fontsize)

    if EV is not None or EC is not None or EF is not None:
        axis.legend(fontsize = legend_fontsize, loc = legendloc)

    axis.tick_params(labelsize = fontsize)


