import os
import sys
import re
#import getenc
#import chardet
from chardet.universaldetector import UniversalDetector

from tklib.tkobject import tkObject
from tklib.tkutils import is_file
import tklib.tkre as tkre

#=========================
# test parameters
path = 'test.ini'
#=========================


def get_encoding(path, defval = 'utf-8'):
    try:
        f = open(path, 'rb')
    except:
        return None

    detector = UniversalDetector()
    for line in f:
        detector.feed(line)
        if detector.done:
            break

    f.close()
    detector.close()

    encoding = detector.result['encoding']
    if encoding == 'SHIFT_JIS' or encoding == 'Windows-1254':
        return 'CP932'
    if encoding is None:
        return defval

    return encoding

def open_chardet(path = None, mode = None, encoding = None, def_encoding = None):
    if encoding is None:
       encoding = get_encoding(path)
#    print("44 encoding=", encoding)
    
    if encoding:
        return open(path, 'r', encoding = encoding)
    else:
        return open(path, 'r', encoding = 'shift_jis')

class tkFile(tkObject):
    def __init__(self, path = None, mode = 'r', encoding = 'utf-8', OpenFile = 1, IsPrint = True, **args):
#        super().__init__(path, mode, **args)
        self.fp   = None
        self.path = None
        self.mode = 'r'
        self.encoding   = encoding
        self.charcode   = None
        self.path       = self.IfYes(path is not None, path, self.path)
        self.mode       = self.IfYes(mode is not None, mode, self.mode)
        self.last_match = None
        self.update(**args)

        if OpenFile and self.path is not None:
            self.open(self.path, self.mode, encoding = encoding, IsPrint = IsPrint)

    def __del__(self):
        self.close()
        super(tkFile, self).__del__()

    def __str__(self):
        return self.ClassPath()

    def SetPath(self, path = None, mode = None):
        self.path = self.IfYes(path is not None, path, self.path)
        self.mode = self.IfYes(mode is not None, mode, self.mode)

    def open(self, path = None, mode = None, IsPrint = True, encoding = None):
        if path is None:
            path = self.path
        if path is None:
            if IsPrint:
                print("")
                print(f"Error in tkfile.open(): path is not given")
            return None

        if mode is None:
            mode = 'r'

        self.SetPath(path, mode)
        if self.fp:
            self.close()

        self.fp = None
#        if not is_file(path):
#            return None

        if encoding is None:
            encoding = self.get('encoding', None)
        if encoding is not None and type(encoding) is not str:
            encoding = None

        if is_file(path) and encoding is None:
            with open(path, 'rb') as f:
                detector = UniversalDetector()
                for line in f:
                    detector.feed(line)
                    if detector.done:
                        break
                detector.close()
                result = detector.result
                if result['encoding'] == 'SHIFT_JIS':
                    encoding = 'CP932'

        if encoding is None:
            encoding = 'utf-8'
#            self.charcode = getenc.getEncode(path)
            self.charcode = 'utf-8'

        try:
            self.fp = open(path, self.mode, encoding = encoding)
        except:
            if not IsPrint:
                return None

            if self.mode == 'r':
#                self.errormsg(sys._getframe().f_code.co_name, 
                self.errormsg("tkfile.open()", 
                    "Can not read [{}]".format(self.path))
            elif self.mode == 'w':
#                self.errormsg(sys._getframe().f_code.co_name, 
                self.errormsg("tkfile.open()", 
                    "Can not write to [{}]".format(self.path))
            else:
#                self.errormsg(sys._getframe().f_code.co_name, 
                self.errormsg("tkfile.open()", 
                    "Can not open [{}]".format(self.path))
            return None

        return self.fp

    def Open(self, path = None, mode = None, IsPrint = True, encoding = None):
        return self.open(path, mode, IsPrint, encoding = encoding)

    def close(self):
        if self.fp is not None:
            self.fp.close()
            self.fp = None

    def Close(self):
        self.close()

    def read(self):
        return self.fp.read()

    def Read(self):
        return self.read()

    def readline(self):
        return self.fp.readline()

        if self.fp is None:
            return None
        try:
            return self.fp.readline()
        except:
            return ''

    def ReadLine(self):
        return self.readline()

    def ReadLineList(self):
        if self.fp is None:
            return None
        self.lines = self.fp.readlines()
        return self.lines

    def ReadCharList(self):
        if self.fp is None:
            return None
        s = self.Read()
        return list(s)

    def ReadLines(self):
        if self.fp is None:
            return None
        return ''.join(self.fp.readlines())

    def write(self, s):
        if self.fp is None:
            return False
        self.fp.write(s)
        return True

    def Write(self, s):
        return self.write(s)

    def write_lines(self, lines):
        if self.fp is None:
            return False

        for l in lines:
            ret = self.fp.write(l)
            if not ret:
                return ret

        return True

    def WriteLines(self, lines):
        return self.write_lines(lines)

    def CopyUntil(self, out, reg = None):
#        print("reg=", reg)
        if reg is None:
            line = self.ReadLine()
            while line:
                print("l=", line)
                out.Write(line)
                line = self.ReadLine()
                return None

        line = self.ReadLine()
        while line:
#            print("line [{}]".format(line), "  reg=[{}]".format(reg))
            if line is None:
                break
            match = re.search(reg, line)
            if match:
                print("hit")
                return match
            out.Write(line)
            line = self.ReadLine()
        return None

# eof can be checked by:  line = fp.readline(); if not line
# or if len(fp.read(1)) == 0 
    def eof(self):
        if self.fp is None:
            return True
        if self.fp.tell() == os.fstat(self.fp.fileno()).st_size:
            return True
        return False

    def Rewind(self):
        if self.fp is None:
            return
        self.fp.seek(0)

    # os.SEEK_SET os.SEEK_CUR  os.SEEK_END
    def Seek(self, offset, whence = os.SEEK_SET):
        if self.fp is None:
            return
        self.fp.seek(offset, whence)

    def Tell(self):
        if self.fp is None:
            return 0
        return self.fp.tell()

    def skip_to(self, pattern, origin = None, re_flag = None):
        if re_flag is None:
            re_flag = re.IGNORECASE
#        print("pattern=[{}]".format(pattern))

        if origin is not None:
            self.fp.seek(origin)

        line = ''
        while 1:
            line = self.ReadLine()
            if not line:
                break

#            print("  line = ", line)
            m = re.search(pattern, line, re_flag)
#            print("[{}] in [{}]".format("[Boot]", line), "  m=", m)
            if m:
                self.last_match = m
                line = self.DelCRLF(line)
                return line
            continue

        return None
    def SkipTo(self, pattern, origin = None, re_flag = None):
        return self.skip_to(pattern, origin = origin, re_flag = re_flag)

    def search_to(self, pattern, origin = None, defval = None):
        re_flag = 'i'
#        re_flag = re.IGNORECASE
#        print("pattern=[{}]".format(pattern))

        if pattern is None:
            return defval

        if origin is not None:
            self.fp.seek(origin)

        line = ''
        while 1:
            line = self.ReadLine()
            if not line:
                break

#            print("  line = ", line)
#            m = re.search(pattern, line, re_flag)
#            print("[{}] in [{}]".format("[Boot]", line), "  m=", m)
            m = tkre.Search(pattern, line, re_flag)
            if m:
                return m
            continue

        if defval is not None:
            return [line, defval]

        return None

    def SearchTo(self, pattern, origin = None, defval = None):
        return self.search_to(pattern, origin = origin, defval = origin)

    def Match(self, reg, str, flag = '', defval = None):
        if str is None:
            return defval

        return tkre.Match(reg, str, flag)

    def Search(self, reg, str, flag = '', defval = None):
        if str is None:
            return defval

        return tkre.Search(reg, str, flag)

    def ReadBlock(self, end_regexp = r'^\s*$'):
        lines = []
        while 1:
            pos = self.fp.tell()
            line = self.fp.eadline()
            if not line or re.search(end_regexp, line):
                self.fp.seek(pos)
                return lines

            lines.append(line)

    def Split(self, reg, str):
        return tkre.Split(reg, str)

    def shSplit(self, reg, str):
        return tkre.shSplit(reg, str)

    def Sub(self, reg, target, str, n = 0):
        return tkre.Sub(reg, target, str, n)

    def SubN(self, reg, target, str, n = 0):
        return tkre.SubN(reg, target, str, n)

    def DelCRLF(self, str):
        return re.sub(r"[\n\r]+$", '', str)

    def DelSpace(self, str):
        return str.strip()

    def DelTopSpace(self, str):
        return str.lstrip()

    def DelLastSpace(self, str):
        return str.rstrip()


def main(path):
    fp = tkFile(path)
    print("fp=", type(fp.fp))
#    vars = ini.readall()

    fp.Rewind()
    s = fp.ReadLine()
    print("ReadLine=", s)

    fp.Rewind()
    ss = fp.ReadLines()
    print("ReadLines=\n", ss)

    fp.Rewind()
    cs = fp.ReadCharList()
    print("Chars=\n", cs)
    fp.Close()

if __name__ == "__main__":
    main(path)


"""
#===============================================
# 文字コード関係変数
#===============================================
# sjis, euc, jis, noconv, utf8
my $FileSystemCharCode = Deps::FileSystemCharCode();
my $PerlCharCode       = Deps::PerlCharCode();
my $MySQLCharCode      = Deps::MySQLCharCode();
my $WebCharCode        = Deps::WebCharCode();
my $WebCharSet         = Deps::WebCharSet();

#===============================================
# スクリプト大域変数
#===============================================
my $LF = Deps::LF();
my $DirectorySeparator = Deps::DirectorySeparator();
my $OS = Deps::OS();

sub new
{
	my ($module, $path, $mode, $charcode, $IsPrintOut) = @_;

	my $this = {};
	bless $this;
	if(defined $path and defined $mode) {
		return undef unless($this->Open($path, $mode, $charcode, $IsPrintOut));
	}
	$this->SetSuppressCharCodeConversion(0);
	return $this;
}

sub DESTROY
{
	my $this = shift;
	$this->Close();
}

sub SetSuppressCharCodeConversion { my ($this,$f)=@_; return $this->{SuppressCharCodeConversion} = $f; }
sub SuppressCharCodeConversion    { return shift->{SuppressCharCodeConversion}; }
sub SuppressIOBuffering
{
	my ($this, $h) = @_;
	$h = $this->HANDLE() if(!defined $h);
	my $oldfh = select($h);
	$|=1;
	select($oldfh); 
}

sub SetSourceCharCode {
	my ($this, $c) = @_;
	return $this->{SoruceCharCode} = $c;
}
sub SourceCharCode { return shift->{SoruceCharCode}; }

sub read
{
	my ($this, $s, $len, $offset) = @_;
	my $ret = read($this->HANDLE(), $s, $len, $offset);
	$_[1] = $s;
	return $ret;
}
sub sysread
{
	my ($this, $s, $len, $offset) = @_;
	my $ret = sysread($this->HANDLE(), $s, $len, $offset);
	$_[1] = $s;
	return $ret;
}
sub syswrite
{
	my ($this, $s, $len, $offset) = @_;
	return syswrite($this->HANDLE(), $s, $len, $offset);
}
sub stat
{
	my ($this) = @_;
	return stat($this->HANDLE());
}
sub lstat
{
	my ($this) = @_;
	return lstat($this->HANDLE());
}
sub fileno
{
	my ($this) = @_;
	return fileno($this->HANDLE());
}
sub flock
{
	my ($this, $operation) = @_;
	return flock($this->HANDLE(), $operation);
}
sub getc
{
	my ($this) = @_;
	return getc($this->HANDLE());
}
sub truncate
{
	my ($this, $length) = @_;
	return truncate($this->HANDLE(), $length);
}

sub WriteFile
{
	my ($this, $filename, $content, $charcode, $mode) = @_;
	Jcode::convert(\$content, $charcode) if($content and $charcode and !$this->SuppressCharCodeConversion());
	open(JFILEFILE, ">$filename") or return 0;
	binmode(JFILEFILE) if(defined $mode and $mode =~ /b/i);
	print JFILEFILE $content;
	close(JFILEFILE);
	return 1;
}

sub ReadFile
{
	my ($this, $filename, $charcode, $mode) = @_;

	open(JFILEFILE, "<$filename") or return undef;
	binmode(JFILEFILE) if(defined $mode and $mode =~ /b/i);
	my $str = do { local $/; <JFILEFILE> };
	close(JFILEFILE);
	Jcode::convert(\$str, $charcode) if($str and $charcode and !$this->SuppressCharCodeConversion());
	return $str;
}

sub SaveFile
{
	my ($this, $str, $filename, $charcode, $mode) = @_;

	open(JFILEFILE, ">$filename") or return undef;
	binmode(JFILEFILE) if(defined $mode and $mode =~ /b/i);
	print JFILEFILE $str;
	close(JFILEFILE);
	return $str;
}


sub ReadFileToLines
{
	my ($this, $filename, $mode, $regexp, $charcode) = @_;
	if($mode !~ /[rw]/) {
		my @lines = $this->ReadFileToLinesPrev($filename, $charcode, $mode);
		return \@lines;
	}

	$this->Open($filename, $mode, $charcode) or return [];
	my @a;
	my $c = 0;
	while(1) {
		my $line = $this->SkipTo($regexp);
		last if(!defined $line);
		$a[$c++] = $line;
	}
	$this->Close();
	return \@a;
}

sub ReadFileToLinesPrev
{
	my ($this, $filename, $charcode, $mode) = @_;

#print "aa [$filename, $charcode, $mode]\n";
	$this->Open($filename, "rb") or return 0;
	my @lines;
	for(my $i = 0 ; ; $i++) {
		my $l = $this->ReadLine();
		last if(!defined $l);
		$lines[$i] = $l;
#print "l:$l";
	}
	$this->Close();
	return @lines;

	open(JFILEFILE, "<$filename") or return 0;
	binmode(JFILEFILE) if(defined $mode and $mode =~ /b/i);
	my @lines = <JFILEFILE>;
	close(JFILEFILE);
	if($charcode and !$this->SuppressCharCodeConversion()) {
		for(my $i = 0 ; $i < @lines ; $i++) {
			Jcode::convert(\$lines[$i], $charcode) if($lines[$i]);
		}
	}
	return @lines;
}

sub Open
{
	my ($this, $filename, $mode, $charcode, $IsPrintOut) = @_;
	$IsPrintOut = 1 if(!defined $IsPrintOut);
	if($this->{inHANDLE} or $this->{outHANDLE}) {
		$this->Close();
	}

	$mode = 'r' if(!defined $mode or $mode eq '');
#	$charcode = 'noconv' unless($charcode);
	$filename = 'con' if($filename eq '');
	Jcode::convert(\$filename, Deps::OSCharCode());

	my $ret = 1;
	if($filename eq 'con') {
		$this->{inHANDLE}  = STDIN;
		$this->{outHANDLE} = STDOUT;
	}
	else {
		if($mode =~ /w/i or $mode =~ /a/i) {
			$this->{outHANDLE} = gensym() if($mode =~ /w/i);
			$filename = ">$filename" if($mode =~ /w/i);
			$filename = ">>$filename" if($mode =~ /a/i);
			$ret = open($this->{'outHANDLE'}, $filename);
			unless($ret) {
				$filename =~ s/^[\<\>]+//;
				print "Error in JFile::Open: Can not write to [$filename].\n"
					if($IsPrintOut);
				$this->Initialize();
				return $ret;
			}
			binmode($this->{outHANDLE}) if($mode =~ /b/i);
		}
		if($mode =~ /r/i) {
			$this->{'inHANDLE'}  = gensym() if($mode =~ /r/i);
			$filename = "<$filename";
			$ret = open($this->{inHANDLE}, $filename);
			unless($ret) {
				$filename =~ s/^[\<\>]+//;
				print "Error in JFile::Open: Can not read [$filename].\n" if($IsPrintOut);
				$this->Initialize();
				return $ret;
			}
			binmode($this->{inHANDLE}) if($mode =~ /b/i);
		}
	}

	$filename =~ s/[\<\>]//g;
	$this->{FileName} = $filename;
	$this->{Mode}     = $mode;
	$this->{CharCode} = $charcode;

	return $ret;
}

sub Close
{
	my ($this) = @_;
	if($this->{'FileName'} and $this->{'FileName'} eq 'con') {
		$this->Initialize();
		return '';
	}
	close($this->{'inHANDLE'}) if($this->{'inHANDLE'});
	close($this->{'outHANDLE'}) if($this->{'outHANDLE'});
	$this->Initialize();
	return '';
}

sub eof
{
	my $this = shift;
	return eof($this->{'inHANDLE'})  if($this->{'inHANDLE'});
	return eof($this->{'outHANDLE'}) if($this->{'outHANDLE'});
	return 1;
}

sub ReadLine
{
	my ($this, $DelSpace, $code, $UseEncodeModule) = @_;
	my $fh = $this->{'inHANDLE'};
	return undef unless($fh);

	$this->{PrevPos} = $this->tell();
	my $line = <$fh>;
	return undef unless(defined $line);
#print "line0: [$line]";
	Utils::DelSpace($line) if($DelSpace);
#$this->{CharCode} = 'utf-8' if($this->{CharCode} eq '');
#print "C=[$this->{CharCode}]\n";
if($code and $UseEncodeModule) {
	$line = decode($code, $line);
}
else {
#	Jcode::convert(\$line, $this->{CharCode}) if($line and $this->{CharCode} and !$this->SuppressCharCodeConversion());
	Utils::convert(\$line, $this->{CharCode}) if($line and $this->{CharCode} and !$this->SuppressCharCodeConversion());
}

	return $line;
}

sub PreReadLine
{
	my ($this) = @_;
	my $pos = $this->tell();
	my $line = $this->ReadLine();
	$this->seek($pos, 0);
	return $line;
}

sub Write
{
	my ($this, $str) = @_;
	Jcode::convert(\$str, $this->{'CharCode'}) if($str and $this->{'CharCode'} and !$this->SuppressCharCodeConversion());
	my $fh = $this->{'outHANDLE'};
	return undef unless($fh);
	print $fh $str;
	return $str;
}

sub print
{
	my ($this, @args) = @_;
$this->{CharCode} = 'utf8' if(defined $this->{CharCode} and $this->{CharCode} =~ /utf-8/i);
	for(my $i = 0 ; $i < @args ; $i++) {
		next if($args[$i] eq '');
		if($this->{CharCode} and $this->{CharCode} ne 'ascii') {
			if($this->{SourceCharCode} and  $this->{SourceCharCode} ne 'ascii') {
				if($this->{SourceCharCode} ne $this->{CharCode} and !$this->SuppressCharCodeConversion()) {
					Jcode::convert(\$args[$i], $this->{CharCode}, $this->{SourceCharCode});
				}
			}
			else {
				if(!$this->SuppressCharCodeConversion()) {
					Jcode::convert(\$args[$i], $this->{CharCode});
				}
			}
		}
#print "arg$i: $args[$i]";
	}
	my $fh = $this->{outHANDLE};
	print $fh @args;
	return @args;
}

sub printf
{
	my ($this, @args) = @_;
	for(my $i = 0 ; $i < @args ; $i++) {
		Jcode::convert(\$args[$i], $this->{'CharCode'}) 
				if($args[$i] and $this->{'CharCode'} and !$this->SuppressCharCodeConversion());
	}
	my $fh = $this->{'outHANDLE'};
	printf $fh @args;
	return @args;
}

sub SetAuxOutput
{
	my ($this, $aux) = @_;
	$this->{AuxOutput} = $aux;
}

sub mprint
{
	my ($this, @args) = @_;
	if(!defined $this->{AuxOutput}) {
		print(@args);
	}
	else {
		$this->{AuxOutput}->print(@args);
	}
	$this->print(@args);
}

sub mprintf
{
	my ($this, @args) = @_;
	if(defined $this->{AuxOutput}) {
		printf(@args);
	}
	else {
		$this->{AuxOutput}->printf(@args);
	}
	$this->printf(@args);
}


sub tell
{
	my ($this) = @_;
	return tell($this->{'inHANDLE'})  if($this->{'inHANDLE'} ne '');
	return tell($this->{'outHANDLE'}) if($this->{'outHANDLE'} ne '');
}

sub rewind { return shift->seek(0,0); }
sub seek
{
	my ($this, $pos, $origin) = @_;
	return seek($this->{'inHANDLE'}, $pos, $origin)  if($this->{'inHANDLE'} ne '');
	return seek($this->{'outHANDLE'}, $pos, $origin) if($this->{'outHANDLE'} ne '');
}

sub BackALine
{
	my ($this, $origin) = @_;
	$this->seek($this->{PrevPos}, $origin);
}

sub Initialize
{
	my ($this) = @_;
	$this->{'inHANDLE'}  = '';
	$this->{'outHANDLE'} = '';
	$this->{'FileName'}  = '';
	$this->{'Mode'}      = '';
	$this->{'CharCode'}  = '';
}

sub FileName
{
	my ($this) = @_;
	return $this->{'FileName'};
}

sub HANDLE
{
	my ($this) = @_;
	return $this->{'inHANDLE'} if($this->{'inHANDLE'} ne '');
	return $this->{'outHANDLE'} if($this->{'outHANDLE'} ne '');
}

sub inHANDLE
{
	my ($this) = @_;
	return $this->{'inHANDLE'};
}

sub outHANDLE
{
	my ($this) = @_;
	return $this->{'outHANDLE'};
}

sub Mode
{
	my ($this) = @_;
	return $this->{'Mode'};
}

sub CharCode
{
	my ($this) = @_;
	return $this->{'CharCode'};
}

"""
