from tklib.tkobject import tkObject


class tkPlot(tkObject):
    """ 
    Helper for matplotlib and plotly
    """

    def __init__(self, **kwargs):
        self.figure_list = []
        self.axes_list = []

        self.update(**kwargs)

    def initialize(self, **kwargs):
        super().__init__(**kwargs)

#    def __del__(self):
#        print("{} destroyed".format(self.name))

    def __str__(self):
        return "graph plot helper object"

    def add_figure(self, figure):
        self.figure_list.append(figure)

    def add_axis(self, axis_inf):
        self.axes_list.append(axis_inf)

    def mpl_connect(self, event, callback):
        for fig in self.figure_list:
            fig.canvas.mpl_connect(event, callback)

    def get_hit_axes(self, target_axis):
        ax_hit = []
        for i in range(len(self.axes_list)):
            axis = self.axes_list[i]["axis"]
            if target_axis == axis:
                 ax_hit.append(i)
        if len(ax_hit) == 0:
            return None
        return ax_hit

    def find_nearest_data1(self, x, y, xlist, ylist):
        minr2 = 1.0e300
        ihit = None
        for i in range(len(xlist)):
            xi = xlist[i]
            yi = ylist[i]
            r2 = (x - xi)**2 + (y - yi)**2
            if minr2 > r2:
                minr2 = r2
                ihit  = i
        return ihit, minr2

    def find_nearest_data(self, iaxes_hit_list, x, y):
        min_idata = 1000000000
        min_r2    = 1.0e30
        min_iaxis = None
        for iaxis in range(len(iaxes_hit_list)):
            inf       = self.axes_list[iaxis]
#            print("inf=", inf)
#            print("x=", inf["x"])
            idata, r2 = self.find_nearest_data1(x, y, inf["x"], inf["y"])
            print("iaxis,idata,r2=", iaxis, idata, r2)
            if r2 < min_r2:
                min_iaxis = iaxis
                min_r2    = r2
                min_idata = idata

        iaxis    = iaxes_hit_list[min_iaxis]
        axis_inf = self.axes_list[iaxis]
        min_idx  = axis_inf['idx'][min_idata]

        return iaxis, axis_inf, min_idata, min_idx, min_r2

    def print_inf(self):
        print("")
        print("Figure info:")
        for i in range(len(self.figure_list)):
            print(f"  {i}: ", self.figure_list[i])
        print("Axes info:")
        for i in range(len(self.axes_list)):
            inf = self.axes_list[i]
            print(f"  {i}: [{inf['label']}] {inf['axis']}")
            
            