#間違っていても責任はとりません、野元
#Single parabric band modelを使ってLorentz数、Keを計算します。
# -*- coding: utf-8 -*- 

import os
import numpy as np
import scipy.integrate as integrate
import sys
import csv
import pandas as pd
import openpyxl


from tklib.tkapplication import tkApplication
from tklib.tkvariousdata import tkVariousData


#===================================
# physical constants
#===================================
pi   = 3.14159265358979323846
pi2  = 2.0 * pi
h    = 6.6260755e-34    # Js";
hbar = 1.05459e-34      # "Js";
c    = 2.99792458e8     # m/s";
e    = 1.60218e-19      # C";
e0   = 8.854418782e-12; # C<sup>2</sup>N<sup>-1</sup>m<sup>-2</sup>";
kB   = 1.380658e-23     # JK<sup>-1</sup>";
me   = 9.1093897e-31    # kg";
R    = 8.314462618      # J/K/mol
a0   = 5.29177e-11      # m";
kBe = 8.6173303e-5 #(kB/e)

# Global variables
infile  = "20221219Ba3GeO.xlsx"
outfile = None

T_label     = 0
S_label     = 1
sigma_label = 2
kappa_label = 3


nmaxiter = 100
eps = 1.0e-10
a =  1000 #max
b = -100 #min

app = None

#===================================
# Treat arguments
#===================================
def pint(str, defval = None):
    try:
        return int(str)
    except:
        return defval

def pfloat(str, defval = None):
    try:
        return float(str)
    except:
        return defval

def getarg(idx, defval = None):
    try:
        return sys.argv[idx]
    except:
        return defval

def getfloatarg(idx, defval = None):
    try:
        return pfloat(sys.argv[idx])
    except:
        return defval

def getintarg(idx, defval = None):
    try:
        return pint(sys.argv[idx])
    except:
        return defval


infile      = getarg(1, defval = infile)
header, ext = os.path.splitext(infile)
filebody    = os.path.basename(header)
outfile     = f'{header}-out.xlsx'
T_label          = getarg(2, defval = T_label)
S_label          = getarg(3, defval = S_label)
sigma_label      = getarg(4, defval = sigma_label)
kappa_label      = getarg(5, defval = kappa_label)


#===================================
# Other functions
#===================================
def usage(app):
    argv = sys.argv
    print("")
    print("Usage: python {} infile T_label S_label(S/cm) sigma_label(S/cm) kappa_label(W/m/K)".format(argv[0]))
    print("   ex: python {} single {} {} {} {} {}".format(argv[0], 
                    infile, T_label, S_label, sigma_label, kappa_label))

def savecsv(outfile, header, datalist):
    try: 
        print("Write to [{}]".format(outfile))
        f = open(outfile, 'w')
    except:
#    except IOError:
        print("line 98 Error: Can not write to [{}]".format(outfile))
    else:
        fout = csv.writer(f, lineterminator='\n')
        fout.writerow(header)
#        fout.writerows(data)
        for i in range(0, len(datalist[0])):
            a = []
            for j in range(len(datalist)):
                a.append(datalist[j][i])
            fout.writerow(a)
        f.close()

def read_file(fname):
    print("")

    datafile = tkVariousData(infile)
    labels, datalist = datafile.Read_minimum_matrix(close_fp = True, usage = usage)

    return datafile, labels, datalist

#Calculation method
def cal_ke(sigma, L, T):
    return sigma * L * T

def Fr(EF, r):
    result = integrate.quad(lambda E: E**r / (1 + np.exp(E - EF)), 0, np.inf)
    #print("felmi_integral = {}".format(result))
    fel = float(result[0])
    #print(fel)
    return fel

def S_comparison(T, S, sigma, EF):
    diff = S - kBe * (2.0 * Fr(EF, r = 1.0) / Fr(EF, r = 0.0) - EF)
    #print("diff = {}".format(diff))
    return diff

def Lorenz(EF):
    answer = kBe**2 * (3.0 * Fr(EF, r = 2.0) / Fr(EF, r = 0.0) - (2.0 * Fr(EF, r = 1.0) / Fr(EF, r = 0.0))**2)
    return answer

def bisection(T, S, sigma, max, min, err):
    EF = (max + min) / 2.0
    ymin = S_comparison(T, S, sigma, min)
    ymax = S_comparison(T, S, sigma, max)
    yh   = S_comparison(T, S, sigma, EF)
    print(f"{EF=} {ymin=} {ymax=} {yh=} {min=} {max=}")

    if ymin * ymax > 0.0:
        print("line Error: ymin * ymax > 0.0")
        app.terminate(pause = True)

    for i in range(nmaxiter):
        print(f"EF={EF:8.3f} d=({ymin:8.4g} {yh:8.4g} {ymax:8.4g}) min={min:8.3f} max={max:8.3f}")

        if abs(yh) < eps:
#        if abs(yh) < eps or abs(max - min) < eps:
            return EF
        
        if ymin * yh > 0.0:
            min = EF
        else:
            max = EF

        EF = (max + min) / 2.0
        ymin = S_comparison(T, S, sigma, min)
        ymax = S_comparison(T, S, sigma, max)
        yh   = S_comparison(T, S, sigma, EF)

    else:
        print(f"Did not converge in {nmaxiter} iterations.")
        exit()

def main():
    global app
    
    app = tkApplication()
    logfile = app.replace_path(infile)
    print(f"Open logfile [{logfile}]")
    app.redirect(targets = ["stdout", logfile], mode = 'w')

    print("")
    print("input file :", infile)
    print("output file:", outfile)
    print("T_label    :", T_label)
    print("S_label    :", S_label)
    print("sigma_label:", sigma_label)
    print("kappa_label:", kappa_label)
    print("")

    print("Read data from [{}]".format(infile))
    datafile, labels, datalist = read_file(infile)
    T_idx = datafile.FindLabelIndex(T_label, flag = 'i')
    S_idx = datafile.FindLabelIndex(S_label, flag = 'i')
    sigma_idx = datafile.FindLabelIndex(sigma_label, flag = 'i')
    kappa_idx = datafile.FindLabelIndex(kappa_label, flag = 'i')
    print("T index: ", T_idx)
    print("S index: ", S_idx)
    print("sigma index: ", sigma_idx)
    print("kappa index: ", kappa_idx)
    Tlabel, Tlist         = datafile.FindDataArray(T_label, flag = 'i')
    Slabel, Slist         = datafile.FindDataArray(S_label, flag = 'i')
    sigmalabel, sigmalist = datafile.FindDataArray(sigma_label, flag = 'i')
    kappalabel, kappalist = datafile.FindDataArray(kappa_label, flag = 'i')

    print("")
    EFlist = []
    EFkTlist = []
    Llist  = []
    kelist  = []
    for ic in range(len(Tlist)):
        T = Tlist[ic]
        S = Slist[ic]
        sigma = sigmalist[ic]
        kappa = kappalist[ic]
        print(f"{ic:3d}: {T=} {S=} {sigma=} {kappa=}")

        EF = bisection(T, S, sigma, max = a, min = b, err = 1e-10)

        L = Lorenz(EF)
        ke = cal_ke(sigma, L, T)

        print(f"EF/kT = {EF}")
        print(f"L  = {L}")
        print(f"ke = {ke} W/m/K")
        EFlist.append(EF * kB * T / e)
        print("kT/e=", kB * T / e)
        EFkTlist.append(EF)
        Llist.append(L)
        kelist.append(ke)

    df = pd.DataFrame(np.array([Tlist,  Slist,    sigmalist,    kappalist,      EFkTlist, EFlist,   Llist, kelist]).T,
                     columns = ["T(K)", "S(V/K)", "simga(S/m)", "kappa(W/m/K)", "EF/kBT", "EF(eV)", "L",   "kappa_e(W/m/K)"])
    print("")
    print("Save L data to [{}]".format(outfile))
    df.to_excel(outfile, index = False, header = True)

    app.terminate("", usage = usage, pause = True)


if (__name__ == '__main__'):
    main()

